// PIV: ANSI C implementation of the API in SP 800-73, Chapter 6 
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The PIV Middleware Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#ifndef __pivdefs_h__
#define __pivdefs_h__

#include <string.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef _WIN32
  #define DYNAMIC_LIB __declspec(dllexport) 
#else
  #define DYNAMIC_LIB

  typedef const char* LPCSTR;
/*  typedef unsigned long DWORD;
  typedef DWORD* LPDOWRD;*/
#endif

#ifdef __cplusplus
extern "C" {
#endif


typedef unsigned long   PIV_RV;
typedef unsigned char   PIV_Bool;
typedef unsigned char   PIV_Byte;
typedef unsigned char   PIV_UChar8;
typedef char            PIV_Char8;
typedef unsigned short  PIV_UShort16;
typedef short           PIV_Short16;
typedef unsigned long   PIV_ULong32;
typedef long            PIV_Long32;

typedef unsigned long   PIV_CARDHANDLE;

typedef unsigned long StatusWord;

#define PIV_OK                                      0 
#define PIV_CONNECTION_DESCRIPTION_MALFORMED        1
#define PIV_CONNECTION_FAILURE                      2
#define PIV_CONNECTION_LOCKED                       3
//#define PIV_CARD_APPLICATION_NO_LONGER_AVAILABLE    4
#define PIV_INVALID_CARD_HANDLE                     5
#define PIV_CARD_APPLICATION_NOT_FOUND              6
#define PIV_AUTHENTICATOR_MALFORMED                 7
#define PIV_AUTHENTICATION_FAILURE                  8
#define PIV_INVALID_OID                             9
#define PIV_DATA_OBJECT_NOT_FOUND                   10
#define PIV_SECURITY_CONDITIONS_NOT_SATISFIED      11
//#define PIV_NO_CURRENT_CONTEXT                     12
#define PIV_INVALID_KEYREF_OR_ALGORITHM            13
#define PIV_INVALID_KEY_OR_KEYALG_COMBINATION	   13
#define PIV_INSUFFICIENT_BUFFER                    20 //14 <- 800-85A lists it as 14, ConfTesting tool (PIVWrapper has 20!!!
//#define PIV_REFERENCE_DATA_NOT_FOUND               14
#define PIV_INPUT_BYTES_MALFORMED                  15
#define PIV_INSUFFICIENT_CARD_RESOURCE             16
//#define PIV_OFFSET_BEYOND_END_OF_DATA_CONTENT      17
//#define PIV_GENERATED_KEY_PAIR_INCOMPATIBLE_WITH_EXISTING_KEY_PAIR     18
#define PIV_UNSUPPORTED_CRYPTOGRAPHIC_MECHANISM    19
//#define PIV_OTHER_CARD_ERROR                       21
#define PIV_CARD_READER_ERROR                      22
//#define PIV_CONNECTION_ERROR                       23


#define PIV_MAX_ATR_SIZE                 33
#define PIV_MAX_READERNAME              100
#define PIV_MAX_APDUSIZE				260
#define PIV_MAX_BUFFERSIZE              260


#ifdef __cplusplus
}
#endif

#endif
