// pivTest: A sample executable to demonstrate usage of the PIV API.
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The PIV Middleware Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.


#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32
#include <windows.h>
#endif

#ifndef __APPLE__
#include <winscard.h>
#else
#include <PCSC/wintypes.h>
#include <PCSC/winscard.h>
#endif

#include "pivdefs.h"
#include "pivtlp224.h"
#include "pivCardEdge.h"
#include "tlvparse.h"
#include "piv.h"

#ifndef MAX_BUFFER_SIZE
  #define MAX_BUFFER_SIZE 65536
#endif

#define MAX_READERS 20

SCARDHANDLE hCard=0;

#define PIV_TEST_UNKNOWN_COMMAND	-1

#define MD5_OID "\x30\x20\x30\x0C\x06\x08\x2A\x86\x48\x86\xF7\x0D\x02\x05\x05\x00\x04\x10"
#define HASH_DATA "\x9D\x24\xBF\x4F\xC7\x8C\x22\x45\xFC\x08\xCE\x9F\xB7\xAE\xB3\x09"
/**************************************************************************************/
// [BCK] addition for pivtest_crypt():
#define CHALLENGE_HASH "\x0d\x67\xa0\xf5\xa2\x2d\x55\x42\xc2\xab\x15\xc8\x24\x42\x13\xae\xdd\xdc\xc5\xc6\x96\xa0\x53\x79\x77\xf0\xfe\x1f\x35\x75\x39\x42\x26\xb1\xe3\x30\x38\xf4\xe4\x7c\x00\x1f\xd0\x66\x3f\xf9\x6d\xa2\x74\x2e\xdc\xd7\x91\x4a\x05\x65\x3b\x09\x48\x63\xf3\x7e\x8f\xa1\x5c\x6e\x9a\xdf\x45\x93\xe0\x50\x4d\xee\x1f\x2f\x49\xf8\xaa\x2e\x44\x49\x12\x57\x74\xba\x89\x03\xe7\x03\x48\x40\x70\x3d\x0f\x7e\x70\x3b\xe1\xd0\xd9\x5f\xde\x2c\xb4\x9c\xc1\x15\x8c\x4d\x55\x3f\xa2\x87\x53\x42\x0a\x79\xb4\xc8\x75\x40\x9d\xef\xdb\x7d\xaf\x9d"
/**************************************************************************************/


#define AT_KEYEXCHANGE 1
#define FALSE 0

int pivtest_connect() {
	long rv;
	LPSTR readers, chosenReader;
	DWORD readersLen;
	int nReaders=0, readerNumber;
	int iList[MAX_READERS]; /* all readers after MAX_READERS will be ignored*/
	PIV_ULong32 bufferLength, i;
	PIV_Byte connectionDescription[500] = {0x7f, 0x21, 0x04, 0x81, 0x00, 0x90, 0x00};
	PIV_ULong32 cdLength=500;
	PIV_Char8 cmd[512];
	PIV_ULong32 cmdLen;

	bufferLength = 0;
	 
	readers	= NULL;

	rv = pivConnect(0, connectionDescription, &cdLength, &hCard);
	if (rv != PIV_OK) {
		printf ("PC/SC appears to be working, but a list of readers could not be retrieved.\n");
		return rv;
	}
    
	readers = malloc(sizeof(char)*cdLength);
	readersLen = cdLength;
	memcpy (readers, connectionDescription, cdLength);
	
	printf("Reader List - Choose a reader number:\n");
	for (i = 0; i < cdLength - 1; i++) {
		++nReaders;
		printf(" - Reader %02d: %s\n", nReaders, &(readers[i]));
		iList[nReaders] = i;
		while (readers[++i] != 0);
	}
	do {
		fflush(stdin); fflush(stdout);
		printf("\nSelect # ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == ' ') { continue; }

		cmdLen = (PIV_ULong32) strlen(cmd);
		cmd[cmdLen-1] = 0; /* NULL the EOL */
		readerNumber = strtol(cmd, (char**)NULL, 10);
		
		if (readerNumber > nReaders || readerNumber <= 0) {
			printf ("Invalid choice.  Please choose a reader from the list typing it's number.\n");
			continue;
		}
		break;
	} while (1);

	chosenReader = &(readers[iList[readerNumber]]);
	printf("Chosen reader: %s\n", chosenReader);
	printf("Attempting to connect...");
	
	
	connectionDescription[0] = 0x7f;
	connectionDescription[1] = 0x21;
	connectionDescription[2] = (PIV_Byte)strlen(chosenReader)+4;
	connectionDescription[3] = 0x81;
	connectionDescription[4] = (PIV_Byte)strlen(chosenReader);
	memcpy(connectionDescription+5, chosenReader, connectionDescription[4]);
	connectionDescription[5+connectionDescription[4]] = 0x90;
	connectionDescription[5+connectionDescription[4]+1] = 0;

	cdLength = connectionDescription[2] + 2;

	rv = pivConnect(0, connectionDescription, &cdLength, &hCard);

	if (rv != PIV_OK) {
		free(readers);
		if (rv == PIV_CONNECTION_LOCKED)
			printf ("\nCard is locked by another application.\n");
		else 
			printf ("\nConnection to card failed.\n");
		return rv;
	}

	printf ("done.\nExclusive connection formed.\n");
	
	free(readers);
	return 0;
}
int pivtest_disconnect() {
	PIV_RV rv;
	printf ("Disconnecting...");
	rv = SCardHookDisconnect (hCard);
	if (rv != SCARD_S_SUCCESS) {
		printf ("\nDisconnect failed.\n");
		return rv;
	}
	else
		printf ("done.\n");
	return 0;
}
int pivtest_select() {

	PIV_Char8 cmd[512];
	PIV_ULong32 cmdLen;

	PIV_Byte pivAID[11] = {0xA0, 0, 0, 3, 8, 0, 0, 0x10, 0, 1, 0};
	PIV_RV rv;

	PIV_Byte *propertyTemplate;
	PIV_ULong32 templateLength;

	printf ("Choose one of the following options:\n\t1.  Select the PIV Applet (A0 00 00 03 08 00 00 10 00 01 00)\n");

	do {
		fflush(stdin); fflush(stdout);
		printf("\nSelect # ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == '\n') { continue; }
		if (cmd[0] == ' ') { continue; }

		cmdLen = (PIV_ULong32) strlen(cmd);
		cmd[cmdLen-1] = 0; /* NULL the EOL */

		if (strcmp(cmd, "1")==0) {
			printf ("Selecting the PIV Applet...\n");
			rv = pivSelectCardApplication((PIV_CARDHANDLE)hCard, pivAID, 11, NULL, &templateLength);
			if (rv != PIV_INSUFFICIENT_BUFFER) {
				printf ("could not select applet.\n");
				return rv;
			}
			propertyTemplate = malloc (sizeof(PIV_Byte)*templateLength);
			rv = pivSelectCardApplication((PIV_CARDHANDLE)hCard, pivAID, 11, propertyTemplate, &templateLength);

			if (rv != PIV_OK) {
				printf ("could not select applet.\n");
				free(propertyTemplate);
				return rv;
			}

			printf ("Applet successfully selected");

			free(propertyTemplate);
			continue;
		}
		printf ("Invalid selection.  Please choose from the list of options above.");

	} while (strcmp(cmd, "1") !=0);

	return 0;
}

int verifyPIN() {
	PIV_Char8 cmd[512];
	PIV_ULong32 cmdLen;
	PIV_Byte authData[15];
	PIV_Byte authDataLen=15;
	
	PIV_RV rv;
	PIV_ULong32 i;


	authData[0] = 0x67;
	authData[1] = 13;
	
	authData[2] = 0x81;
	authData[3] = 8;
	authData[4] = 0xff;
	authData[5] = 0xff;
	authData[6] = 0xff;
	authData[7] = 0xff;
	authData[8] = 0xff;
	authData[9] = 0xff;
	authData[10]= 0xff;
	authData[11]= 0xff;

	authData[12]= 0x83;
	authData[13]= 1;
	
	printf ("Which PIN would you like to verify?\n");
	printf ("\t1. Global PIN (0x00)\n");
	printf ("\t2. PIV Card Application PIN (0x80)\n");
	
	do {
		fflush(stdin); fflush(stdout);
		printf("PIN # ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == '\n' || cmd[0] == ' ') { printf ("Please choose from the list.\n");continue; }

		cmdLen = (PIV_ULong32) strlen(cmd);
		cmd[cmdLen-1] = 0; /* NULL the EOL */

		if (strcmp(cmd, "1")==0)
			authData[14] = 0;
		else if (strcmp(cmd, "2")==0)
			authData[14] = 0x80;
		else {
			printf ("That selection is not available.  Please choose from the list.\n");
			continue;	
		}
		break;
	} while(1);
	
	printf ("Please enter your PIN [1234]");
	do {
		fflush(stdin); fflush(stdout);
		printf("\nPIN # ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == '\n') { 
			printf("Using 1234...\n");
			authData[4] = '1'; authData[5] = '2';
			authData[6] = '3'; authData[7]= '4';
		}
		else {
			cmdLen = (PIV_ULong32) strlen(cmd);
			cmd[cmdLen-1] = 0; /* NULL the EOL */

			if (cmdLen-1 > 8){
				printf ("The PIN is a maximum of 8 characters long.  Try again\n");
				continue;
			}
			for (i=0; i<cmdLen-1; i++)
				authData[4+i] = cmd[i];
		}
//		for (i=0; i<authDataLen; i++)
//			printf("%02x ", authData[i]);
		printf("\n");
		rv = pivLogIntoCardApplication((PIV_CARDHANDLE)hCard, authData, authDataLen); 
		if (rv != PIV_OK) {
			printf ("Verification failed.\n");
			return rv;
		}
		break;
	} while (1);

	return 0;
}

int xauth() {
	printf ("External Authentication test code has been removed.\n");
	return 0;
}

int pivtest_login() {
	PIV_Char8 cmd[512];
	PIV_ULong32 cmdLen;

	printf ("How would you like to authenticate?\n");
	printf ("\t1.  PIN Verification\n");
	printf ("\t2.  External Authentication (challenge response)\n");

	do {
		fflush(stdin); fflush(stdout);
		printf("\nLogin # ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == '\n') { continue; }
		if (cmd[0] == ' ') { continue; }

		cmdLen = (PIV_ULong32) strlen(cmd);
		cmd[cmdLen-1] = 0; /* NULL the EOL */

		if (strcmp(cmd, "1")==0)
			return verifyPIN();			
		else if (strcmp(cmd, "2")==0)
			return xauth();
		else continue;

		break;
	} while (strcmp(cmd, "back")!=0);

	return 0;
}

int printElementList () {
	printf ("\t1.  Card Capability Container\n");
	printf ("\t2.  Card Holder Unique Identifier\n");
	printf ("\t3.  X.509 Certificate for PIV Authentication\n"); 
	printf ("\t4.  Card Holder Fingerprint I\n");
	printf ("\t5.  Card Holder Fingerprint II\n");
	printf ("\t6.  Printed Information\n");
	printf ("\t7.  Card Holder Facial Image\n"); 
	printf ("\t8.  X.509 Certificate for Digital Signature\n");
	printf ("\t9.  X.509 Certificate for Key Management\n");
	printf ("\t10. X.509 Certificate for Card Authentication\n");
	printf ("\t11. Security Object\n");
	return 0;
}

int pivtest_getdata() {
	PIV_Byte oid[100];
	PIV_ULong32 oidLength;
	PIV_RV rv;
	PIV_Byte *data;
	PIV_ULong32 dataLength;
	PIV_Char8 cmd[512];
	PIV_ULong32 cmdLen;

	printf ("Which data element would you like to get?\n");
	printElementList();

	do {
		fflush(stdin); fflush(stdout);
		printf("\nGet # ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == '\n') { continue; }
		if (cmd[0] == ' ') { continue; }

		cmdLen = (PIV_ULong32) strlen(cmd);
		cmd[cmdLen-1] = 0; /* NULL the EOL */
	 
		if (strcmp(cmd, "1")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.1.219.0", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else if (strcmp(cmd, "2")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.48.0", 25);
			oid[25]=0;
			oidLength = 25;
		}
		else if (strcmp(cmd, "3")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.1.1", 24);
			oid[24]=0;
			oidLength = 24;
		}
		else if (strcmp(cmd, "4")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.96.16", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else if (strcmp(cmd, "5")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.96.17", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else if (strcmp(cmd, "6")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.48.1", 25);
			oid[25]=0;
			oidLength = 25;
		}
		else if (strcmp(cmd, "7")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.96.48", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else if (strcmp(cmd, "8")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.1.0", 24);
			oid[24]=0;
			oidLength = 24;
		}
		else if (strcmp(cmd, "9")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.1.2", 24);
			oid[24]=0;
			oidLength = 24;
		}
		else if (strcmp(cmd, "10")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.5.0", 24);
			oid[24]=0;
			oidLength = 24;
		}
		else if (strcmp(cmd, "11")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.144.0", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else continue;

		rv = pivGetData((PIV_CARDHANDLE)hCard, oid, oidLength, NULL, &dataLength);
		if (rv != PIV_OK && rv != PIV_INSUFFICIENT_BUFFER) //OK for buffer size discovery.
		{
			printf ("Error encountered");
			return rv;
		}
		data = malloc (sizeof(PIV_Byte)*dataLength);
		rv = pivGetData((PIV_CARDHANDLE)hCard, oid, 3, data, &dataLength);

#ifdef _DEBUG
		dumpData(data,dataLength);
#endif
		free(data);

		if (rv != PIV_OK) {
			printf ("Error encountered");
			return rv;
		}

		break;
	} while (1);

	return 0;
}
int pivtest_logout() {
	PIV_RV rv;
	printf ("Logging out of PIV application...");
	rv = pivLogoutOfCardApplication ((PIV_CARDHANDLE)hCard);
	if (rv != PIV_OK) {
		printf ("Failed.\n");
	}
	else 
		printf ("Done.\n");
	return rv;
}

/***************************************************************************
int pivtest_crypt() {
	

//printf("All cryptographic functionality of this example application has been removed.\n");

//return 0;

	PIV_RV rv;
	PIV_Byte algID = 0x06; // Hardcoded for now
	PIV_Byte keyReference;
	int dwKeySpec = AT_KEYEXCHANGE;
	unsigned char pSignature[300];
	int pSignatureLen;
	int ii;

	unsigned char whatever[200];

	whatever[0] = 0x7c;
	whatever[1] = 0x81;
	whatever[2] = 0x85;
	whatever[3] = 0x82;
	whatever[4] = 0x00;
	whatever[5] = 0x81;
	whatever[6] = 0x81;
	whatever[7] = 0x80;
	whatever[8] = 0x00;
	whatever[9] = 0x01;

	memset(&whatever[10], 0xff, 91);
	whatever[101] = 0x00;
	memcpy(&whatever[102], MD5_OID, 18);
	memcpy(&whatever[120], HASH_DATA, 16);
	keyReference = 0x9a;
	pSignatureLen = 200;

	rv = pivCrypt((PIV_CARDHANDLE)hCard, algID, keyReference,whatever, 136, pSignature, &pSignatureLen);
	if (rv != PIV_OK) {
		printf("Crypt Operation failed!!!!#####@@@@@@. %d\n", rv);
		return FALSE;
	}

	else {
		printf("\nReturned data length is ->%d\n", pSignatureLen);
		for (ii=0; ii < pSignatureLen; ii++) {
			printf("%02X ", pSignature[ii]);
		}
	}
	return 0;
}
***************************************************************************/
int pivtest_crypt() {
 
//printf("All cryptographic functionality of this example application has been removed.\n");
//return 0;
 
 PIV_RV rv;
 PIV_Byte algID = 0x06; // Hardcoded for now
 PIV_Byte keyReference;
 int dwKeySpec = AT_KEYEXCHANGE;
 unsigned char pSignature[300];
 int pSignatureLen;
 int ii;
 
 unsigned char whatever[200];
 
 /* [BCK] old header
 whatever[0] = 0x7c;
 whatever[1] = 0x81;
 whatever[2] = 0x85;
 whatever[3] = 0x82;
 whatever[4] = 0x00;
 whatever[5] = 0x81;
 whatever[6] = 0x81;
 whatever[7] = 0x80;
 whatever[8] = 0x00;
 whatever[9] = 0x01;
 */
 
 // [BCK] new header: Dyn Auth Tag, 133 bytes total, Challenge, 128 bytes
 whatever[0] = 0x7c;
 whatever[1] = 0x81;
 whatever[2] = 0x85;
 whatever[3] = 0x81;
 whatever[4] = 0x81; 
 whatever[5] = 0x80;
 
 /* [BCK] old challenge
 memset(&whatever[10], 0xff, 91);
 whatever[101] = 0x00;
 memcpy(&whatever[102], MD5_OID, 18);
 memcpy(&whatever[120], HASH_DATA, 16);
 */
 
 // [BCK] new challenge (128 bytes)
 memcpy(&whatever[6], CHALLENGE_HASH, 128);
 
 // [BCK] new footer: request response, presenting zero bytes, need one byte per template standard (0x00)
 whatever[134] = 0x82;
 whatever[135] = 0x00;
 whatever[136] = 0x00;

/* int pindex;
 for(pindex = 134; pindex < 136; pindex++)
 {
  printf("%i ",whatever[pindex]);
 }*/

 keyReference = 0x9a;
 pSignatureLen = 200;
 
 rv = pivCrypt((PIV_CARDHANDLE)hCard, algID, keyReference,whatever, 136, pSignature, &pSignatureLen);
 if (rv != PIV_OK) {
  printf("Crypt Operation failed!!!!#####@@@@@@. %d\n", rv);
  return FALSE;
 }
 
 else {
  printf("\nReturned data length is ->%d\n", pSignatureLen);
  for (ii=0; ii < pSignatureLen; ii++) {
   printf("%02X ", pSignature[ii]);
  }
 }
 return 0;
}

/**************************************************************************/

int pivtest_putdata() {
	PIV_Byte oid[100];
	PIV_ULong32 oidLength;
	PIV_RV rv;
	PIV_Byte data[15000];
	PIV_ULong32 dataLength;
	PIV_Char8 cmd[512];
	PIV_ULong32 cmdLen;

	printf ("Which data element would you like to write to?\n");
	printElementList();

	do {
		fflush(stdin); fflush(stdout);
		printf("\nPut # ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == '\n') { continue; }
		if (cmd[0] == ' ') { continue; }

		cmdLen = (PIV_ULong32) strlen(cmd);
		cmd[cmdLen-1] = 0; /* NULL the EOL */

		if (strcmp(cmd, "1")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.1.219.0", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else if (strcmp(cmd, "2")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.48.0", 25);
			oid[25]=0;
			oidLength = 25;
		}
		else if (strcmp(cmd, "3")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.1.1", 24);
			oid[24]=0;
			oidLength = 24;
		}
		else if (strcmp(cmd, "4")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.96.16", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else if (strcmp(cmd, "5")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.96.17", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else if (strcmp(cmd, "6")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.48.1", 25);
			oid[25]=0;
			oidLength = 25;
		}
		else if (strcmp(cmd, "7")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.96.48", 26);
			oid[26]=0;
			oidLength = 26;
		}
		else if (strcmp(cmd, "8")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.1.0", 24);
			oid[24]=0;
			oidLength = 24;
		}
		else if (strcmp(cmd, "9")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.1.2", 24);
			oid[24]=0;
			oidLength = 24;
		}
		else if (strcmp(cmd, "10")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.5.0", 24);
			oid[24]=0;
			oidLength = 24;
		}
		else if (strcmp(cmd, "11")==0) {
			memcpy (oid, "2.16.840.1.101.3.7.2.144.0", 26);
			oid[26]=0;
			oidLength = 26;
		}	
		else continue;

		dataLength = 850;
		memset(data, (PIV_Byte)0xcb, dataLength);

		rv = pivPutData((PIV_CARDHANDLE)hCard, oid, 3, data, dataLength);
		if (rv != PIV_OK) {
			printf ("Error encountered");
			return rv;
		}
		break;

	}while(strcmp(cmd, "back") != 0);

	return 0;
}
int pivtest_generate() {

	PIV_Char8 cmd[512];
	PIV_ULong32 cmdLen;
	
	PIV_Byte pivAID[11] = {0xA0, 0, 0, 3, 8, 0, 0, 0x10, 0, 1, 0};
	StatusWord sw;

	PIV_Byte reference;

	PIV_Byte *keyObjects;
	PIV_ULong32 keyObjectsLength=250;

	
	keyObjects = malloc (sizeof(PIV_Byte)*keyObjectsLength);

	printf ("Which Key would you like to generate?\n");
	printf ("\t1. PIV Authentication Key (0x9A)\n");
	printf ("\t2. Card Application Administration Key (0x9B)\n");
	printf ("\t3. Digital Signature Key (0x9C)\n");
	printf ("\t4. Key Management Key (0x9D)\n");

	do {
		fflush(stdin); fflush(stdout);
		printf("\nGenerate # ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == '\n') { continue; }
		if (cmd[0] == ' ') { continue; }

		cmdLen = (PIV_ULong32) strlen(cmd);
		cmd[cmdLen-1] = 0; /* NULL the EOL */

		if (strcmp(cmd, "1")==0)
			reference = 0x9A;
		else if (strcmp(cmd, "2")==0)
			reference = 0x9B;
		else if (strcmp(cmd, "3")==0)
			reference = 0x9C;
		else if (strcmp(cmd, "4")==0)
			reference = 0x9D;
		else continue;

		
        sw = pivGenerateKeyPair((PIV_CARDHANDLE)hCard, reference, PIV_CMID_RSA1024,  NULL, &keyObjectsLength);
		break;
	} while (1);

	free(keyObjects);
	return 0;
}

int executeCommand(int command) {
	switch (command) {
		case PIV_CONNECT:
			return pivtest_connect();
		case PIV_DISCONNECT:
			return pivtest_disconnect();
		case PIV_SELECT_CARD_APPLICATION:
			return pivtest_select();
		case PIV_LOG_IN_TO_CARD_APPLICATION:
			return pivtest_login();
		case PIV_GET_DATA:
			return pivtest_getdata();
		case PIV_LOG_OUT_OF_CARD_APPLICATION:
			return pivtest_logout();
		case PIV_CRYPT:
			return pivtest_crypt();
		case PIV_PUT_DATA:
			return pivtest_putdata();
		case PIV_GENERATE_KEY_PAIR:
			return pivtest_generate();
		default:
			return PIV_TEST_UNKNOWN_COMMAND;
	}
}

int printHelp() {
	printf("----------------------------------------------------\n");
	printf("OpCode:\t\tDescription:\n");
    printf("     %2d\t\t- Connect\n", PIV_CONNECT);
	printf("     %2d\t\t- Disconnect\n", PIV_DISCONNECT);
	printf("     %2d\t\t- Select Card Application\n", PIV_SELECT_CARD_APPLICATION);
	printf("     %2d\t\t- Log in to Card Application\n", PIV_LOG_IN_TO_CARD_APPLICATION);
	printf("     %2d\t\t- Get Data\n", PIV_GET_DATA);
	printf("     %2d\t\t- Log out of Card Application\n", PIV_LOG_OUT_OF_CARD_APPLICATION);
	printf("     %2d\t\t- Crypto functions\n", PIV_CRYPT);
	printf("     %2d\t\t- Put Data\n", PIV_PUT_DATA);
	printf("     %2d\t\t- Generate Key Pair\n", PIV_GENERATE_KEY_PAIR);
	printf("     %2d\t\t- Set javacard for use\n", PIV_OC_HACK);
#ifdef USE_OPENSSL
#endif
	printf("\n----------------------------------------------------\n");
	printf("      q\t\t- Quit\n");
	printf("   help\t\t- Display this menu\n");
	printf("\n");

	return 0;
}

int invokeMenu() {
	PIV_Char8 cmd[512];
	PIV_Char8 *arg1;
	PIV_ULong32 cmdLen;

	int command, rv;

	printf("\nPIV Command Tool - type \"help\" for help\n");

	do {
		fflush(stdin); fflush(stdout);

		printf("\nPIV Menu> ");

		if (fgets(cmd, sizeof(cmd), stdin) == NULL) { continue; }
		if (cmd[0] == '\n') { continue; }
		if (cmd[0] == ' ') { continue; }

		cmdLen = (PIV_ULong32) strlen(cmd);
		cmd[cmdLen-1] = 0; /* NULL the EOL */

		arg1 = strtok(cmd, " ");

		if (strcmp(arg1, "help") == 0) {
			printHelp();
		} else if (strcmp(arg1, "quit") == 0 || strcmp(arg1, "q") == 0 || strcmp(arg1, "exit") == 0 ||
			strcmp(arg1, "bye") == 0) {
				pivtest_logout();
				pivtest_disconnect();
				printf("Bye\n");
		} else {
			command = atoi(arg1);
			rv = executeCommand(command);
			if (rv == PIV_OK)
				continue;
            if (rv == PIV_TEST_UNKNOWN_COMMAND)
				printf("Unknown command. Type \"help\" for help\n");
		}
	} while (strcmp(cmd, "quit") != 0 && strcmp(cmd, "q") != 0 && strcmp(cmd, "exit") != 0 &&
			strcmp(cmd, "bye") != 0);
	
	return 0;
}
	
int main() {
	//setServerURL("piv.mobile-mind.com");
	//setServerPort(9025);
	return invokeMenu();
}
