// Security System
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include <Security_System.h>

// Include CryptoPP headers
#pragma warning(disable:4267)
//  #include "aes.h"
#include "asn.h"
//  #include "cbcmac.h"
//  #include "channels.h"
#include "des.h"
#include "dh.h"
//  #include "dsa.h"
#include "ec2n.h"
#include "eccrypto.h"
#include "ecp.h"
//  #include "files.h"
//  #include "fips140.h"
#include "hex.h"
//  #include "hmac.h"
#include "modes.h"
//  #include "mqueue.h"
#include "nbtheory.h"
#include "oids.h"
#include "osrng.h"
//  #include "pkcspad.h"
#include "randpool.h"
#include "rng.h"
#include "rsa.h"
//  #include "sha.h"
//  #include "skipjack.h"
//  #include "trdlocal.h"

USING_NAMESPACE(CryptoPP)


static RandomPool rng ( Security_System & the_security_system )
  {
    Octet_String seed(8);

    the_security_system . generate_cryptogram ( seed , 8 ) ;

    RandomPool pool ;

    pool . Put ( seed . data ( ) , 8 ) ;

    return pool ;
  }

static Integer convert_to_Integer ( const Octet_String & octet_string )
  {
    Integer i ;
    i . Decode ( octet_string. data ( ) , octet_string . length ( ) ) ;
    return i ;
  }

void Security_System::Triple_DES_ECB_128_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    ECB_Mode<DES_EDE2>::Encryption encryption;
    encryption . SetKey ( key . data ( ) , key . length ( ) ) ;
    
    Octet input_buffer [ 8 ] ;
    Octet output_buffer [ 8 ] ;

    Octet_String::cpy ( input_buffer  , sizeof ( input_buffer  ) , plaintext  ) ;

    encryption.ProcessBlocks( output_buffer, input_buffer, 1 ) ;

    Octet_String::cpy ( cyphertext , output_buffer , sizeof ( output_buffer ) ) ;
  } ;


void Security_System::Triple_DES_ECB_128_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    ECB_Mode<DES_EDE2>::Decryption decryption;
    decryption . SetKey ( key . data ( ) , key . length ( ) ) ;
    
    Octet input_buffer [ 8 ] ;
    Octet output_buffer [ 8 ] ;

    Octet_String::cpy ( input_buffer  , sizeof ( input_buffer  ) , cyphertext  ) ;

    decryption.ProcessBlocks( output_buffer, input_buffer, 1 ) ;

    Octet_String::cpy ( plaintext , output_buffer , sizeof ( output_buffer ) ) ;
  } ;


void Security_System::Triple_DES_CBC_128_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    CBC_Mode<DES_EDE2>::Encryption encryption;
    encryption . SetKey ( key . data ( ) , key . length ( ) ) ;
    
    Octet input_buffer [ 8 ] ;
    Octet output_buffer [ 8 ] ;

    Octet_String::cpy ( input_buffer  , sizeof ( input_buffer  ) , plaintext  ) ;

    encryption.ProcessBlocks( output_buffer, input_buffer, 1 ) ;

    Octet_String::cpy ( cyphertext , output_buffer , sizeof ( output_buffer ) ) ;
  } ;


void Security_System::Triple_DES_CBC_128_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    CBC_Mode<DES_EDE2>::Decryption decryption;
    decryption . SetKey ( key . data ( ) , key . length ( ) ) ;
    
    Octet input_buffer [ 8 ] ;
    Octet output_buffer [ 8 ] ;

    Octet_String::cpy ( input_buffer  , sizeof ( input_buffer  ) , cyphertext  ) ;

    decryption.ProcessBlocks( output_buffer, input_buffer, 1 ) ;

    Octet_String::cpy ( plaintext , output_buffer , sizeof ( output_buffer ) ) ;
  } ;

void Security_System::Triple_DES_ECB_192_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    ECB_Mode<DES_EDE3>::Encryption encryption;
    encryption . SetKey ( key . data ( ) , key . length ( ) ) ;
    
    Octet input_buffer [ 8 ] ;
    Octet output_buffer [ 8 ] ;

    Octet_String::cpy ( input_buffer  , sizeof ( input_buffer  ) , plaintext  ) ;

    encryption.ProcessBlocks( output_buffer, input_buffer, 1 ) ;

    Octet_String::cpy ( cyphertext , output_buffer , sizeof ( output_buffer ) ) ;
  } ;


void Security_System::Triple_DES_ECB_192_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    ECB_Mode<DES_EDE3>::Decryption decryption;
    decryption . SetKey ( key . data ( ) , key . length ( ) ) ;
    
    Octet input_buffer [ 8 ] ;
    Octet output_buffer [ 8 ] ;

    Octet_String::cpy ( input_buffer  , sizeof ( input_buffer  ) , cyphertext  ) ;

    decryption.ProcessBlocks( output_buffer, input_buffer, 1 ) ;

    Octet_String::cpy ( plaintext , output_buffer , sizeof ( output_buffer ) ) ;
  } ;

void Security_System::Triple_DES_CBC_192_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    CBC_Mode<DES_EDE3>::Encryption encryption;
    encryption . SetKey ( key . data ( ) , key . length ( ) ) ;
    
    Octet input_buffer [ 8 ] ;
    Octet output_buffer [ 8 ] ;

    Octet_String::cpy ( input_buffer  , sizeof ( input_buffer  ) , plaintext  ) ;

    encryption.ProcessBlocks( output_buffer, input_buffer, 1 ) ;

    Octet_String::cpy ( cyphertext , output_buffer , sizeof ( output_buffer ) ) ;
  } ;


void Security_System::Triple_DES_CBC_192_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    CBC_Mode<DES_EDE3>::Decryption decryption;
    decryption . SetKey ( key . data ( ) , key . length ( ) ) ;
    
    Octet input_buffer [ 8 ] ;
    Octet output_buffer [ 8 ] ;

    Octet_String::cpy ( input_buffer  , sizeof ( input_buffer  ) , cyphertext  ) ;

    decryption.ProcessBlocks( output_buffer, input_buffer, 1 ) ;

    Octet_String::cpy ( plaintext , output_buffer , sizeof ( output_buffer ) ) ;
  } ;

void Security_System::RSA_1024_encrypt ( RSA_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    RSA::PrivateKey privKey;

	privKey.Initialize ( convert_to_Integer ( key . m ) ,
                         convert_to_Integer ( key . e ) ,
                         convert_to_Integer ( key . d ) ) ;

    privKey . CalculateInverse( rng ( * this ), Integer ( const_cast < Octet * > ( plaintext . data ( ) ), plaintext . length ( ) ) ) .  
              Encode ( const_cast < Octet * > ( cyphertext . data ( ) ),  privKey . GetModulus ( ) . ByteCount ( ) );
  } ;


void Security_System::RSA_1024_decrypt ( RSA_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    RSA::PublicKey pubKey;

    pubKey.Initialize ( convert_to_Integer ( key . m ) ,
                        convert_to_Integer ( key . e ) ) ;

    pubKey . ApplyRandomizedFunction( rng ( * this ), Integer ( const_cast < Octet * > ( cyphertext . data ( ) ), cyphertext . length ( ) ) ) .  
             Encode(const_cast < Octet * > ( plaintext . data ( ) ),  pubKey . GetModulus ( ) . ByteCount ( ) );
  } ;


void Security_System::RSA_2048_encrypt ( RSA_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    RSA::PrivateKey privKey;

	privKey.Initialize ( convert_to_Integer ( key . m ) ,
                         convert_to_Integer ( key . e ) ,
                         convert_to_Integer ( key . d ) ) ;

    privKey . CalculateInverse( rng ( * this ), Integer ( const_cast < Octet * > ( plaintext . data ( ) ), plaintext . length ( ) ) ) .  
              Encode ( const_cast < Octet * > ( cyphertext . data ( ) ),  privKey . GetModulus ( ) . ByteCount ( ) );
  } ;


void Security_System::RSA_2048_decrypt ( RSA_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    RSA::PublicKey pubKey;

    pubKey.Initialize ( convert_to_Integer ( key . m ) ,
                        convert_to_Integer ( key . e ) ) ;

    pubKey . ApplyRandomizedFunction( rng ( * this ), Integer ( const_cast < Octet * > ( cyphertext . data ( ) ), cyphertext . length ( ) ) ) .  
             Encode(const_cast < Octet * > ( plaintext . data ( ) ),  pubKey . GetModulus ( ) . ByteCount ( ) );
  } ;

void Security_System::RSA_3072_encrypt ( RSA_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    RSA::PrivateKey privKey;

	privKey.Initialize ( convert_to_Integer ( key . m ) ,
                         convert_to_Integer ( key . e ) ,
                         convert_to_Integer ( key . d ) ) ;

    privKey . CalculateInverse( rng ( * this ), Integer ( const_cast < Octet * > ( plaintext . data ( ) ), plaintext . length ( ) ) ) .  
              Encode ( const_cast < Octet * > ( cyphertext . data ( ) ),  privKey . GetModulus ( ) . ByteCount ( ) );
  } ;


void Security_System::RSA_3072_decrypt ( RSA_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    RSA::PublicKey pubKey;

    pubKey.Initialize ( convert_to_Integer ( key . m ) ,
                        convert_to_Integer ( key . e ) ) ;

    pubKey . ApplyRandomizedFunction( rng ( * this ), Integer ( const_cast < Octet * > ( cyphertext . data ( ) ), cyphertext . length ( ) ) ) .  
             Encode(const_cast < Octet * > ( plaintext . data ( ) ),  pubKey . GetModulus ( ) . ByteCount ( ) );
  } ;

void Security_System::ECC_P_224_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::secp224k1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    encryptor . Encrypt ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          plaintext . length ( ) ,
                          cyphertext . data ( ) ) ;
  } ;


void Security_System::ECC_P_224_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::secp224k1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    DecodingResult result = 
    decryptor . Decrypt  ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          cyphertext . length ( ) ,
                          plaintext . data ( ) ) ;

    if ( ! result . isValidCoding )
      throw RAPDU_Packet::SW_VERIFICATION_FAILED ;
  } ;


void Security_System::ECC_K_233_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::sect233k1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    encryptor . Encrypt ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          plaintext . length ( ) ,
                          cyphertext . data ( ) ) ;
  } ;


void Security_System::ECC_K_233_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::sect233k1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    DecodingResult result = 
    decryptor . Decrypt  ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          cyphertext . length ( ) ,
                          plaintext . data ( ) ) ;

    if ( ! result . isValidCoding )
      throw RAPDU_Packet::SW_VERIFICATION_FAILED ;
  } ;

void Security_System::ECC_B_233_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::sect233r1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    encryptor . Encrypt ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          plaintext . length ( ) ,
                          cyphertext . data ( ) ) ;
  } ;


void Security_System::ECC_B_233_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::sect233r1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    DecodingResult result = 
    decryptor . Decrypt  ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          cyphertext . length ( ) ,
                          plaintext . data ( ) ) ;

    if ( ! result . isValidCoding )
      throw RAPDU_Packet::SW_VERIFICATION_FAILED ;
  };


void Security_System::ECC_P_256_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::secp256k1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    encryptor . Encrypt ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          plaintext . length ( ) ,
                          cyphertext . data ( ) ) ;
  } ;


void Security_System::ECC_P_256_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::secp256k1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    DecodingResult result = 
    decryptor . Decrypt  ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          cyphertext . length ( ) ,
                          plaintext . data ( ) ) ;

    if ( ! result . isValidCoding )
      throw RAPDU_Packet::SW_VERIFICATION_FAILED ;
  } ;



void Security_System::ECC_K_283_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::sect283k1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    encryptor . Encrypt ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          plaintext . length ( ) ,
                          cyphertext . data ( ) ) ;
  } ;


void Security_System::ECC_K_283_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::sect283k1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    DecodingResult result = 
    decryptor . Decrypt  ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          cyphertext . length ( ) ,
                          plaintext . data ( ) ) ;

    if ( ! result . isValidCoding )
      throw RAPDU_Packet::SW_VERIFICATION_FAILED ;
  } ;

void Security_System::ECC_B_283_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::sect283r1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    encryptor . Encrypt ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          plaintext . length ( ) ,
                          cyphertext . data ( ) ) ;
  } ;


void Security_System::ECC_B_283_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    DL_PrivateKey_EC<ECP> privKey;

    Integer privateKeyExponent ( convert_to_Integer ( key . p ) ) ;

    privKey.Initialize( ASN1::sect283r1() , privateKeyExponent );

    ECIES<ECP>::Decryptor decryptor ( privKey ) ;
	ECIES<ECP>::Encryptor encryptor ( decryptor );

    DecodingResult result = 
    decryptor . Decrypt  ( rng ( * this ) , 
                          const_cast < Octet * > ( plaintext . data ( ) ) ,
                          cyphertext . length ( ) ,
                          plaintext . data ( ) ) ;

    if ( ! result . isValidCoding )
      throw RAPDU_Packet::SW_VERIFICATION_FAILED ;
  } ;

void Security_System::AES_ECB_128_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;


void Security_System::AES_ECB_128_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;


void Security_System::AES_CBC_128_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;


void Security_System::AES_CBC_128_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;


void Security_System::AES_ECB_192_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;

void Security_System::AES_ECB_192_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;


void Security_System::AES_CBC_192_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;


void Security_System::AES_CBC_192_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;



void Security_System::AES_ECB_256_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;


void Security_System::AES_ECB_256_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;



void Security_System::AES_CBC_256_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;


void Security_System::AES_CBC_256_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext )
  {
    key ;  // TODO: unused
    plaintext ;  // TODO: unused
    cyphertext ;  // TODO: unused
    throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
  } ;

RSA_Key_Pointer Security_System::Generate_RSA_Key_Pair( const unsigned int bits , const unsigned int publicExponent  )
{
    AutoSeededRandomPool rng;
    InvertibleRSAFunction keyPair;
    Integer exponent( publicExponent );

    keyPair.Initialize( dynamic_cast<RandomNumberGenerator&>(rng), bits , exponent );

    Integer n = keyPair . GetModulus( ) ;
    Octets::Pointer nbytes ( Octets::alloc ( n . ByteCount())) ;
    n . Encode( nbytes -> body( ) , n . ByteCount( ) ) ;

    Integer d = keyPair . GetPrivateExponent( ) ;
    Octets::Pointer dbytes ( Octets::alloc ( d . ByteCount())) ;
    d . Encode( dbytes -> body( ) , d . ByteCount( ) ) ;

    Integer e = keyPair . GetPublicExponent( ) ;
    Octets::Pointer ebytes ( Octets::alloc ( e . ByteCount())) ;
    e . Encode( ebytes -> body( ) , e . ByteCount( ) ) ;

    RSA_Key_Pointer the_RSA_key_pointer =
      RSA_Key_Pointer::object_pointer_to_indexed_pointer 
        ( new RSA_Key ( nbytes -> body ( ) , n . ByteCount( ) , 
                        dbytes -> body ( ) , d . ByteCount( ),  
                        ebytes -> body ( ) , e . ByteCount( ) ) );

    Octets::free ( nbytes ) ;
    Octets::free ( dbytes ) ;
    Octets::free ( ebytes ) ;

    return the_RSA_key_pointer ;
}

ECC_Key_Pointer Security_System::Generate_ECC_Key_Pair( const Octet mechanism )
{
    AutoSeededRandomPool rng;
	DL_PrivateKey_EC<ECP> privkey;
	DL_PublicKey_EC<CryptoPP::ECP> pubkey;

    switch(mechanism)
    {
    case Algorithm_Identifier::ECC_P_224:
	    privkey.Initialize(rng, ASN1::secp224k1());
        break;
    case Algorithm_Identifier::ECC_K_233:
	    privkey.Initialize(rng, ASN1::sect233k1());
        break;
    case Algorithm_Identifier::ECC_B_233:
	    privkey.Initialize(rng, ASN1::sect233r1());
        break;
    case Algorithm_Identifier::ECC_P_256:
	    privkey.Initialize(rng, ASN1::secp256k1());
        break;
    case Algorithm_Identifier::ECC_K_283:
	    privkey.Initialize(rng, ASN1::sect283k1());
        break;
    case Algorithm_Identifier::ECC_B_283:
	    privkey.Initialize(rng, ASN1::sect283r1());
        break;
    default: break;
        throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
    }

    Integer p = privkey.GetPrivateExponent();

    Octets::Pointer pbytes ( Octets::alloc ( p . ByteCount())) ;
    p . Encode( pbytes -> body( ) , p . ByteCount( ) ) ;

	privkey.MakePublicKey(pubkey);

    ECPPoint point = pubkey.GetPublicElement();
    Integer x = point.x;
    Integer y = point.y;

    Octets::Pointer xbytes ( Octets::alloc ( x . ByteCount())) ;
    x . Encode( xbytes -> body( ) , x . ByteCount( ) ) ;

    Octets::Pointer ybytes ( Octets::alloc ( y . ByteCount())) ;
    y . Encode( ybytes -> body( ) , y . ByteCount( ) ) ;

    ECC_Key_Pointer the_ECC_key_pointer =
      ECC_Key_Pointer::object_pointer_to_indexed_pointer 
        ( new ECC_Key ( pbytes -> body ( ) , p . ByteCount( ) , 
                        xbytes -> body ( ) , x . ByteCount( ),  
                        ybytes -> body ( ) , y . ByteCount( ) ) );

    Octets::free ( pbytes ) ;
    Octets::free ( xbytes ) ;
    Octets::free ( ybytes ) ;

    return the_ECC_key_pointer ;
}


bool Security_Attribute::authorized ( const Security_System & security_system , 
                                      Octet access_mode )
const
{
  Access_Rule_Pointer_Sequence::const_iterator p;

  if ( the_access_rule_pointers . begin ( ) == the_access_rule_pointers . end ( ) )
    // There are no access mode/security condition pairs 
    return false;

  bool some_rule_matches = false ;

  for ( p = the_access_rule_pointers . begin ( ) ; 
        p != the_access_rule_pointers . end ( ) ; 
        p ++ )
    {
      const Access_Rule & this_access_rule  = * * p ;
      if ( this_access_rule . mode_matches ( access_mode ) )
        {
          some_rule_matches = true ;

          if ( ! this_access_rule . evaluate ( security_system ) )
            return false ;
      }
    }
  return some_rule_matches ;
}

bool Security_Condition_Data_Object_Sequence::evaluate_status
      ( const Security_System & security_system ) const
  {
    // The PIV Reference Implementation supports key reference access rules look as follows:
    // { ACR_SC_DO : 
    //   { ACR_SC_KEY_REFERENCE   : <key number> } , 
    //   { ACR_SC_USAGE_QUALIFIER : ACR_UQ_USE } }
    Security_Condition_Data_Object_Sequence::const_iterator p = begin ( ) ;
    if ( p == end ( ) )
      return false ;
    Primitive_Security_Condition_Data_Object_Pointer key_reference_data_object_pointer
      ( * p ++ ) ;
    if ( p == end ( ) )
      return false ;
    Primitive_Security_Condition_Data_Object_Pointer usage_qualifier_data_object_pointer
      ( * p ++ ) ;
    if ( p != end ( ) )
      return false ;
    if ( key_reference_data_object_pointer -> tag ( ) == BER_TLV_Tag ( ACR_SC_KEY_REFERENCE )
          && 1 == key_reference_data_object_pointer -> length ( ) . value ( ) 
          && usage_qualifier_data_object_pointer -> tag ( ) == BER_TLV_Tag ( ACR_SC_USAGE_QUALIFIER )
          && 1 == usage_qualifier_data_object_pointer -> length ( ) . value ( ) ) 
      {
        Key_Reference key_reference ( key_reference_data_object_pointer -> values ( ) [ 0 ] ) ;
        
        Octet usage_qualifier = usage_qualifier_data_object_pointer -> values ( ) [ 0 ] ;
        const Key_Entry_Pointer & the_key_entry_pointer = security_system . the_key_store . get_key_entry_pointer ( key_reference ) ;
        switch ( the_key_entry_pointer . type ( ) )
          {
            case Key_Entry_Pointer::generic:
              throw RAPDU_Packet::SW_SECURITY_STATUS_NOT_SATISFIED ;
              break ;

            case Key_Entry_Pointer::PIN:
              if ( ACR_UQ_USE != usage_qualifier )
                throw RAPDU_Packet::SW_SECURITY_STATUS_NOT_SATISFIED ;
              break ;

            case Key_Entry_Pointer::symmetric_key:
            if ( ACR_UQ_EXTERNAL_AUTHENTICATION != usage_qualifier )
                throw RAPDU_Packet::SW_SECURITY_STATUS_NOT_SATISFIED ;
                break ;

            case Key_Entry_Pointer::RSA_key:
              break ;

            default:
              throw RAPDU_Packet::SW_SECURITY_STATUS_NOT_SATISFIED ;
          }
        if ( security_system . the_security_status . authenticated ( key_reference ) )
          return true ;
      }
    return false ;
  }


bool Security_Condition_Data_Object_Sequence::evaluate_AND
      ( const Security_System & security_system ) const
  {
    Security_Condition_Data_Object_Sequence::const_iterator p = begin ( ) ;

    // Although the usual extension of AND to a sequence of booleans defines
    // AND(<empty sequence>) == true in order to perserve
    // AND(<a concatenated with b>) == AND(AND(<a>),AND(<b>))
    // the PIV Reference Implementation follows the more restrictive 
    // smart card industry approach and defines AND(<empty sequence>)==false
    if ( p == end ( ) )
      return false ;

    do
      {
        switch ( Security_Condition_Data_Object_Pointer( * p ) . get_type ( ) )
          {
            case Security_Condition_Data_Object_Pointer::primitive:
              if ( ! Primitive_Security_Condition_Data_Object_Pointer ( * p ) -> evaluate ( security_system ) )
                return false ;
              break ;

            case Security_Condition_Data_Object_Pointer::constructed:
              if ( ! Constructed_Security_Condition_Data_Object_Pointer ( * p ) -> evaluate ( security_system ) )
                return false ;
              break ;

            default:
              return false ;
          }
        
      }
    while ( ( ++ p ) != end ( ) ) ;

    return true ;
  }


bool Security_Condition_Data_Object_Sequence::evaluate_OR
      ( const Security_System & security_system ) const
  {
    for ( Security_Condition_Data_Object_Sequence::const_iterator p = begin ( ) ;
          p != end ( ) ;
          p ++ )
      switch ( Security_Condition_Data_Object_Pointer( * p ) . get_type ( ) )
        {
          case Security_Condition_Data_Object_Pointer::primitive:
            if ( Primitive_Security_Condition_Data_Object_Pointer ( * p ) -> evaluate ( security_system ) )
              return true ;
            break ;

          case Security_Condition_Data_Object_Pointer::constructed:
            if ( Constructed_Security_Condition_Data_Object_Pointer ( * p ) -> evaluate ( security_system ) )
              return true ;
            break ;

          default:
            return false ;
        }
    return false ;
  }


bool Security_Condition_Data_Object_Sequence::evaluate_NOT
      ( const Security_System & security_system ) const
  {
        Security_Condition_Data_Object_Sequence::const_iterator p = begin ( ) ;

    // This is implemented as AND above, except that the security conditions should
    // all evaluate as false
    if ( p == end ( ) )
      return false ;

    do
      {
        switch ( Security_Condition_Data_Object_Pointer( * p ) . get_type ( ) )
          {
            case Security_Condition_Data_Object_Pointer::primitive:
              if ( Primitive_Security_Condition_Data_Object_Pointer ( * p ) -> evaluate ( security_system ) )
                return false ;
              break ;

            case Security_Condition_Data_Object_Pointer::constructed:
              if ( Constructed_Security_Condition_Data_Object_Pointer ( * p ) -> evaluate ( security_system ) )
                return false ;
              break ;

            default:
              return false ;
          }
        
      }
    while ( ( ++ p ) != end ( ) ) ;

    return true ;
  }
      
      
Security_Condition_Data_Object_Pointer 
Security_Condition_Data_Object::create ( const BER_TLV_Pointer & data_object_pointer )
  {
    return 
        data_object_pointer -> tag ( ) . denotes_constructed_TLV ( ) 
          ? Security_Condition_Data_Object_Pointer 
              ( new Constructed_Security_Condition_Data_Object 
                      ( * Constructed_Security_Condition_Data_Object_Pointer 
                            ( data_object_pointer ) ) )
          : Security_Condition_Data_Object_Pointer 
              ( new Primitive_Security_Condition_Data_Object   
                      ( * Primitive_Security_Condition_Data_Object_Pointer   
                            ( data_object_pointer ) ) ) ;
  } ;

void Security_Condition_Data_Object_Pointer::attach ( void * the_object_pointer )
  { add_reference_to_referenced_object ( the_object_pointer  ) ; } ;


void Security_Condition_Data_Object_Pointer::detach ( void * the_object_pointer )
  {
    if ( remove_reference_to_referenced_object ( the_object_pointer ) )
      {
        // Simulate a virtual destructor
        switch ( get_type ( ) )
          {
            case generic:
              throw Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ( ) ;
              break ;

            case primitive:
              reinterpret_cast < Primitive_Security_Condition_Data_Object * > ( the_object_pointer )
                -> ~Primitive_Security_Condition_Data_Object ( ) ;
              break ;

            case constructed:
              reinterpret_cast < Constructed_Security_Condition_Data_Object * > ( the_object_pointer )
                -> ~Constructed_Security_Condition_Data_Object ( ) ;
              break ;
          } ;
        deallocate_referenced_object ( the_object_pointer ) ;
      }
  } ;



// The routines below deal with the choice of hierarchy.
//
// A Security_Condition_Data_Object is a BER_TLV
// But is a Primitive_Security_Condition_Data_Object 
//   a Security_Condition_Data_Object or
//   a Primitive_BER_TLV?
// And similarly for Constructed_Security_Condition_Data_Object.
// Rather than work out the multiple inheritance hierarchy,
// the choice is that a Primitive_Security_Condition_Data_Object 
// is a Primitive_BER_TLV and the assumption is made that
// Security_Condition_Data_Object has no instance variables.
// Hence the outermost reinterpret_cast in the following pointer hacking 
// is correct.

const Security_Condition_Data_Object * Security_Condition_Data_Object_Pointer::operator -> ( ) const
  throw ( Wrong_Security_Condition_Data_Object_Pointer_Type_Exception )
  {
    switch ( the_type )
      {
      case generic:
        throw Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ( ) ;

      case primitive:
        return
          reinterpret_cast < const Security_Condition_Data_Object * > 
              ( static_cast < const BER_TLV * >
                    ( reinterpret_cast < const Primitive_Security_Condition_Data_Object * > 
                        ( object_pointer ( ) ) ) ) ;

      case constructed:
        return
          reinterpret_cast < const Security_Condition_Data_Object * > 
              ( static_cast < const BER_TLV * >
                  ( reinterpret_cast < const Constructed_Security_Condition_Data_Object * > 
                      ( object_pointer ( ) ) ) ) ;
      }
    return 0 ;
  } ;

Security_Condition_Data_Object * Security_Condition_Data_Object_Pointer::operator -> ( )       
  throw ( Wrong_Security_Condition_Data_Object_Pointer_Type_Exception )
  {
    switch ( the_type )
      {
      case generic:
        throw Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ( ) ;

      case primitive:
        return
          reinterpret_cast <       Security_Condition_Data_Object * > 
              ( static_cast <       BER_TLV * >
                    ( reinterpret_cast <       Primitive_Security_Condition_Data_Object * > 
                        ( object_pointer ( ) ) ) ) ;

      case constructed:
        return
          reinterpret_cast <       Security_Condition_Data_Object * > 
              ( static_cast <       BER_TLV * >
                  ( reinterpret_cast <       Constructed_Security_Condition_Data_Object * > 
                      ( object_pointer ( ) ) ) ) ;
      }
    return 0 ;
  } ;

// Debugging
void Security_System::dump ( )
  {
    for ( Octet key_number = 1 ; 
          key_number < Key_Reference::number_of_key_references ; 
          key_number ++ )
      {
        Key_Entry_Pointer the_key_entry_pointer = 
          the_key_store . get_key_entry_pointer ( key_number ) ;
        switch ( the_key_entry_pointer . type ( ) )
          {
            case Key_Entry_Pointer::generic:
              if ( 3 < the_verbosity ) 
                debug_message ( "Key[%2d]  : <none>\n" , key_number ) ;
              break ;

            case Key_Entry_Pointer::PIN:
              debug_message ( "Key[%2d] %c :" , key_number ,
                              the_security_status.authenticated(key_number)
                                ? 'Y'
                                : 'n' ) ;
              {
                const PIN_Entry & the_PIN_entry = 
                  * PIN_Entry_Pointer ( the_key_entry_pointer ) ; 
                debug_message (" PIN: retries=%d remaining=%d :  " ,
                              the_PIN_entry . default_retries ,
                              the_PIN_entry . retries_remaining ) ;
                for ( size_t i = 0 ; i < the_PIN_entry . PIN_pointer -> length ( ) ; i ++ )
                  debug_message ( " 0x%02X" , the_PIN_entry . PIN_pointer -> get_octet ( i ) ) ;
              }
              debug_message ( "\n" ) ;
              break ;

            case Key_Entry_Pointer::symmetric_key:
              debug_message ( "Key[%2d] %c :" , key_number ,
                              the_security_status.authenticated(key_number)
                                ? 'Y'
                                : 'n' ) ;
              {
                const Symmetric_Key_Entry & the_key_entry = 
                  * Symmetric_Key_Entry_Pointer ( the_key_entry_pointer ) ; 
                debug_message (" Symmetric_Key: " ) ;
                for ( size_t i = 0 ; i < the_key_entry . key_pointer ->  length ( ) ; i ++ )
                  debug_message ( " 0x%02X" , the_key_entry . key_pointer -> get_octet ( i ) ) ;
              }
              debug_message ( "\n" ) ;
              break ;

            case Key_Entry_Pointer::RSA_key:
              debug_message ( "Key[%2d] %c :" , key_number ,
                              the_security_status.authenticated(key_number)
                                ? 'Y'
                                : 'n' ) ;
              {
                const RSA_Key_Entry & the_key_entry = 
                  * RSA_Key_Entry_Pointer ( the_key_entry_pointer ) ; 
                debug_message (" RSA_Key:  n: " ) ;
                for ( size_t i = 0 ; i < the_key_entry . key_pointer -> m . length ( ) ; i ++ )
                  debug_message ( " 0x%02X" , the_key_entry . key_pointer -> m . get_octet ( i ) ) ;
                debug_message ("\n                      d: " ) ;
                for ( size_t i = 0 ; i < the_key_entry . key_pointer -> d . length ( ) ; i ++ )
                  debug_message ( " 0x%02X" , the_key_entry . key_pointer -> d . get_octet ( i ) ) ;
              }
              debug_message ( "\n" ) ;
              break ;

            default:
              debug_message ( "Key[%2d] ???\n" , key_number ) ;
          }
      }
  } ;

