// TLV
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include "TLV.h"

void BER_TLV_Pointer::detach ( void * the_object_pointer )
  {
    if ( remove_reference_to_referenced_object ( the_object_pointer ) )
      {
        // Simulate a virtual destructor
        switch ( get_type ( ) )
          {
            case generic:
              throw Wrong_BER_TLV_Pointer_Type_Exception ( ) ;
              break ;

            case primitive:
              reinterpret_cast < Primitive_BER_TLV * > ( the_object_pointer )
                -> ~Primitive_BER_TLV ( ) ;
              break ;

            case constructed:
              reinterpret_cast < Constructed_BER_TLV * > ( the_object_pointer )
                -> ~Constructed_BER_TLV ( ) ;
              break ;
          } ;
        deallocate_referenced_object ( the_object_pointer ) ;
      }
  } ;

BER_TLV_Pointer 
BER_TLV::create ( const Octet_String & octet_string , Memory::offset_t start )
  {
    return 
        BER_TLV_Tag ( octet_string , start ) . denotes_constructed_TLV ( ) 
          ? BER_TLV_Pointer ( new Constructed_BER_TLV ( octet_string , start ) )
          : BER_TLV_Pointer ( new Primitive_BER_TLV   ( octet_string , start ) ) ;
  } ;

BER_TLV_Pointer
BER_TLV::create ( const Octet_String & octet_string , Memory::offset_t start , Memory::offset_t stop )
  {
    return 
        BER_TLV_Tag ( octet_string , start ) . denotes_constructed_TLV ( ) 
          ? BER_TLV_Pointer ( new Constructed_BER_TLV ( octet_string , start , stop ) )
          : BER_TLV_Pointer ( new Primitive_BER_TLV   ( octet_string , start , stop ) ) ;
  } ;


string BER_TLV_Sequence::spell ( ) const
  {
    string s ;
    if ( begin ( ) == end ( ) )
    {
        s = "<none>";
    }
    else
    {
        for ( iterator p = begin ( ) ; p != end ( ) ; p ++ )
        {
          BER_TLV_Pointer bp = *p ;
          string bs ;
          switch ( bp . get_type ( ) )
            {
              case BER_TLV_Pointer::generic:
                throw Wrong_BER_TLV_Pointer_Type_Exception ( ) ;

              case BER_TLV_Pointer::primitive:
                bs = Primitive_BER_TLV_Pointer ( bp ) -> spell ( ) ;
                break ;

              case BER_TLV_Pointer::constructed:
                bs = Constructed_BER_TLV_Pointer ( bp ) -> spell ( ) ;
                break ;
            }
          s += " " ;
          s += bs ;
        }
    }
    return s ;
  } ;

const BER_TLV * BER_TLV_Pointer::operator -> ( ) const
  throw ( Wrong_BER_TLV_Pointer_Type_Exception )
  {
    switch ( the_type )
      {
      case generic:
        throw Wrong_BER_TLV_Pointer_Type_Exception ( ) ;

      case primitive:
        return
          reinterpret_cast < const Primitive_BER_TLV * > 
              ( object_pointer ( ) ) ;

      case constructed:
        return
          reinterpret_cast < const Constructed_BER_TLV * > 
              ( object_pointer ( ) ) ;
      }
    return 0 ;
  } ;

BER_TLV * BER_TLV_Pointer::operator -> ( )       
  throw ( Wrong_BER_TLV_Pointer_Type_Exception )
  {
    switch ( the_type )
      {
      case generic:
        throw Wrong_BER_TLV_Pointer_Type_Exception ( ) ;

      case primitive:
        return
          reinterpret_cast <       Primitive_BER_TLV * > 
              ( object_pointer ( ) ) ;

      case constructed:
        return
          reinterpret_cast <       Constructed_BER_TLV * > 
              ( object_pointer ( ) ) ;
      }
    return 0 ;
  } ;
