// Application - Generic Application for the Reference Implementation Framework
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__APPLICATION_H_INCLUDED__)
#define __APPLICATION_H_INCLUDED__


#if defined(_MSC_VER )
  #if defined(APPLICATION_EXPORTS)
    #define APPLICATION_DLL _declspec(dllexport)
  #else //  defined ( APPLICATION_EXPORTS )
    #define APPLICATION_DLL _declspec(dllimport)
  #endif  // defined ( APPLICATION_EXPORTS )
#else
   #define APPLICATION_DLL
#endif

class APPLICATION_DLL File_Manager ;
#include <File_Manager.h>

class APPLICATION_DLL Application 
{

public:
  friend class File_Manager ;

  Application( File_Manager * file_manager_pointer ,
                const File_System_Pointer & file_system_pointer ,
                const Security_System_Pointer & security_system_pointer ,
                Octet_String_Pointer & challenge_pointer )
  : the_file_manager_pointer ( file_manager_pointer ) ,
    the_file_system_pointer ( file_system_pointer ) ,
    the_security_system_pointer ( security_system_pointer ) ,
    the_challenge_pointer ( challenge_pointer )
    { } ;

  virtual long initialize      ( int v ) ;
  virtual long exchange_APDUs   ( const CAPDU_Packet & command , 
                                        RAPDU_Packet & response )    = 0 ;

  virtual long finalize        ( void )  ;

  virtual long create_runnable_instance ( const unsigned long flags ,
                                          const Octet_String & module_AID ,
                                          const Octet_String & requested_instance_AID ,
                                          const Octet_String & install_parameters ,
                                          Octet_String & created_instance_AID ) ;
  
  virtual long connect_runnable_instance ( const Octet_String & module_AID ) ;

  virtual const Octet_String & AID ( ) = 0 ;

protected:
  File_Manager * the_file_manager_pointer ;
  File_System_Pointer the_file_system_pointer ;
  Security_System_Pointer the_security_system_pointer ;
  Octet_String_Pointer & the_challenge_pointer ;
};

typedef Application *  
Application_Creator ( File_Manager * file_manager_pointer ,
                      const File_System_Pointer & file_system_pointer ,
                      const Security_System_Pointer & security_system_pointer ,
                      Octet_String_Pointer & challenge_pointer ) ;
#endif  // __APPLICATION_H_INCLUDED__
