// Card Manager
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__CARD_MANAGER_H_INCLUDED__)
#define __CARD_MANAGER_H_INCLUDED__

#include <Manager.h>
#include "Card_Registry.h"

// #include "Card_Service.h"
class Card_Service ; // forward declaration to avoid circular include dependency


// INSTALL APDU constants
#define P1_INSTALL_FOR_LOAD                        0x02
#define P2_INSTALL_FOR_LOAD                        0x00
#define P1_INSTALL_FOR_INSTALL_AND_MAKE_SELECTABLE 0xC0
#define P2_INSTALL_FOR_INSTALL_AND_MAKE_SELECTABLE 0x00
#define P1_INSTALL_FOR_MAKE_SELECTABLE             0x80
#define P2_INSTALL_FOR_MAKE_SELECTABLE             0x00

#define LOAD_FILE_FORMAT_JAVACARD_CAP              0x00
#define LOAD_FILE_FORMAT_WFSC_RTE                  0x01
#define LOAD_FILE_FORMAT_WIN32_DLL                 0x02

//
// Runtime Environments and Virtual Machines
//
#define TARGET_ARCHITECTURE_NATIVE_CODE            0x00

#define TARGET_ARCHITECTURE_MULTOS_MEL_BYTECODE    0x01
#define TARGET_ARCHITECTURE_JAVA_CARD_BYTECODE     0x02
#define TARGET_ARCHITECTURE_WFSC_RTE_BYTECODE      0x03
#define TARGET_ARCHITECTURE_7816_FILESYSTEM        0x04
#define TARGET_ARCHITECTURE_CLI_BYTECODE           0x05
#define TARGET_ARCHITECTURE_PASCAL_P_CODE          0x06
#define TARGET_ARCHITECTURE_BCPL_O_CODE            0x07

// LOAD APDU constants
#define P1_LOAD_MORE_BLOCKS                        0x00
#define P1_LOAD_LAST_BLOCK                         0x80
#define TAG_LOAD_FILE_DATA_BLOCK                   0xc4

static Octet truncated_PIV_AID [ ] = { 0xA0 , 0x00 , 0x00 , 0x03 , 0x08 , 0x00 , 0x00 , 0x10 , 0x00 } ;

static Octet full_V_01_PIV_AID [ ] = { 0xA0 , 0x00 , 0x00 , 0x03 , 0x08 , 0x00 , 0x00 , 0x10 , 0x00 , 0x01, 0x00 } ;

class Card_Manager: public Manager
{
public:
  Card_Manager ( ) ;
  
  // The APDU Server interface
  virtual long
  initialize ( int v ) ;
  
  virtual APDU_Handler 
  exchange_APDUs ;

  virtual long
  finalize ( void ) ;

  virtual long
  power_up ( void ) ;

  virtual long
  power_down ( void ) ;

  virtual void 
  get_ATR ( unsigned char * ATR , 
            unsigned int & ATR_length ) ;

  // the Card_Service API
  unsigned char
  get_card_content_state ( Card_Service * caller ) ;
  
  bool
  set_card_content_state ( Card_Service * caller ,
                           unsigned char new_state ) ;
  
  unsigned char
  get_card_state ( Card_Service * caller ) ;
  
  bool
  lock_card ( Card_Service * caller ) ;
  
  bool
  terminate_card ( Card_Service * caller ) ;
  
  bool
  set_ATR_hist_bytes ( Card_Service * caller ,
                       Octet_String & new_hist_bytes ) ;

  // Persistent state below
  Card_Registry_Pointer   the_registry_pointer ;
  // Persistent state above

  bool store_persistent_state ( FILE * the_persistent_storage_output_file , 
                                unsigned short & checksum ) ;

  bool load_persistent_state ( FILE * the_persistent_storage_input_file , 
                               unsigned short & checksum ) ;

private:

  long
  install_for_load ( Octet_String the_payload ) ;

  long
  install_for_install_and_make_selectable  ( Octet_String the_payload ) ;

  long
  install_for_make_selectable  ( Octet_String the_payload ) ;

  void add_preloaded_applets_to_registry ( ) ;

  Application_Manager *   managers [ Registry_Entry :: manager_invalid + 1 ];
  Registry_Entry::Manager active_manager_index;
  Octet                   current_AID [ MAX_AID_LENGTH ] ;
  unsigned int            current_AID_length ;
  Octet                   ISD_AID [ MAX_AID_LENGTH ] ;
  unsigned int            ISD_AID_length ;
  Octet_String            pending_load_file_AID ;
} ;

#endif  //  __CARD_MANAGER_H_INCLUDED__