/* file tlpclib-debug.c
 *
 * NIST PIV Demo
 *
 * Copyright  2005 Electrosoft Services Inc.
 * http://www.electrosoft-inc.com
 * 
 * This file codes and documents all the necessary debug functions.  It was
 * from csp-debug.c
 * 
 * Adapted from work of Romain Chantereau <rchantereau@entrouvert.com> (CSP11 code)
 */
#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include "tlpclib-debug.h"

static HANDLE g_debugFileHandle = INVALID_HANDLE_VALUE; /**< Handle to the debug
                                                           log file.*/
void TlpclibCloseDebug()
{
    if(g_debugFileHandle == INVALID_HANDLE_VALUE)
    {
        CloseHandle(g_debugFileHandle);
    }
}

void TlpclibFdebug(const char *message)
{
    DWORD   bytesWritten;
    DWORD   dwPos;

    // If the log file handle is invalid...
    if(g_debugFileHandle == INVALID_HANDLE_VALUE)
    {
        // Try to get a valid file handle.
        g_debugFileHandle = CreateFile(TEXT(TLPCLIB_DEBUG_FILE), // The debug file
                                     GENERIC_WRITE,              // Open for writing
                                     FILE_SHARE_READ,            // Allow multiple readers
                                     NULL,                       // No security
                                     OPEN_ALWAYS,                // If not exists, create
                                     FILE_ATTRIBUTE_NORMAL,      // normal file
                                     NULL);				         // No attribute template
    }
    // If the log file handle is still invalid, error
    if(g_debugFileHandle == INVALID_HANDLE_VALUE)
        return;

    // Go to the end of the file
    dwPos = SetFilePointer(g_debugFileHandle, 0, NULL, FILE_END);
    // Lock at the end of file, on length of message +1
    if(!LockFile(g_debugFileHandle,dwPos,0, (DWORD) strlen(message),0))
        return;

    // Write the message to the file
    if(!WriteFile(g_debugFileHandle, message, (DWORD) strlen(message), &bytesWritten, NULL))
    {
        UnlockFile(g_debugFileHandle, dwPos, 0, (DWORD) strlen(message),0);
        return;
    }

    // Unlock the file
    UnlockFile(g_debugFileHandle, dwPos, 0, (DWORD) strlen(message),0);
}

void TlpclibFdebug2(const char *message, int len)
{
    DWORD   bytesWritten;
    DWORD   dwPos;

    // If the log file handle is invalid...
    if(g_debugFileHandle == INVALID_HANDLE_VALUE)
    {
        // Try to get a valid file handle
        g_debugFileHandle = CreateFile(TEXT(TLPCLIB_DEBUG_FILE), // The debug file
                                     GENERIC_WRITE,              // Open for writing
                                     FILE_SHARE_READ,            // Allow multiple readers
                                     NULL,                       // No security
                                     OPEN_ALWAYS,                // If not exists, create
                                     FILE_ATTRIBUTE_NORMAL,      // normal file
                                     NULL);                      // No attribute template
    }
    // If the log file handle is still invalid, error
    if(g_debugFileHandle == INVALID_HANDLE_VALUE)
        return;

    // Go to the end of the file
    dwPos = SetFilePointer(g_debugFileHandle, 0, NULL, FILE_END);
    // Lock at the end of file, on length of message +1
    if(!LockFile(g_debugFileHandle,dwPos,0, (DWORD) strlen(message),0))
        return;

    // Write the message to the file
    if(!WriteFile(g_debugFileHandle, message, len, &bytesWritten, NULL))
    {
        UnlockFile(g_debugFileHandle, dwPos, 0, (DWORD) strlen(message),0);
        return;
    }
    // Unlock the file
    UnlockFile(g_debugFileHandle, dwPos, 0, (DWORD) strlen(message),0);
}