// Interface to PC/SC
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include "tlpclib-debug.h"
#include "tlpclib.h"
#include <Storage.h>
#include <iostream>
#include <APDU_Packet.h>
#include <TLP_APDU_Server.h>
#include <windows.h>

TLP_APDU_Server the_tlp_apdu_server;
CAPDU_Packet cAPDU;
RAPDU_Packet rAPDU;

int TLPCLIB_SCardConnect(char *hostAddress, int portNumber)
{
	int verbosity = 0;

	//TLPCLIB_DEBUG_MSG(2, "------------- TLPCLIB SCardConnect -------------\n");

	the_tlp_apdu_server.set_verbosity ( verbosity ) ;

	if ( 0 != the_tlp_apdu_server.set_hostname ( hostAddress ) ) {

		//TLPCLIB_DEBUG_MSG(2, "set hostname failed\n");
		std::cout << "set_hostname failed." << endl;
		return -1;
	}

	the_tlp_apdu_server.set_port ( portNumber );
	the_tlp_apdu_server.report_initialization_errors = true;

	if ( 0 != the_tlp_apdu_server.initialize ( verbosity ) ) {

		//TLPCLIB_DEBUG_MSG(2, "initialize failed\n");
		std::cout << "initialize failed." << endl;
		return -1;
	}

	//TLPCLIB_DEBUG_MSG(2, "Powering up cref...\n");

#ifdef TLPCLIB_DEBUG
	std::cout << "Powering up cref.\n";
#endif

	if ( 0 != the_tlp_apdu_server.power_up() ) {

		//TLPCLIB_DEBUG_MSG(2, "Power up failed\n");
		std::cout << "power_up failed." << endl;
		return -1;
	}

	//TLPCLIB_DEBUG_MSG(2, "SCardConnect succeeded!\n");

#ifdef TLPCLIB_DEBUG
	std::cout << " calling the_tlp_apdu_server.get_hostname() -- returns:" << the_tlp_apdu_server.get_hostname() << "\n";
	std::cout << " calling the_tlp_apdu_server.get_port() -- returns:" << the_tlp_apdu_server.get_port() << "\n";
#endif

	return 0;
}

int TLPCLIB_SCardDisconnect()
{
#ifdef TLPCLIB_DEBUG
	std::cout << "Powering down cref.\n";
#endif

	if ( 0 != the_tlp_apdu_server.power_down() )
    {
		std::cout << "power_down failed." << endl;
      return -1;
    };
    
  if ( 0 != the_tlp_apdu_server.finalize () )
    {
		std::cout << "finalize failed." << endl;
      return -1;
    };

  return 0;
}

int TLPCLIB_SCardTransmit(unsigned char *txBuffer, unsigned long txLength,
					   unsigned char *rxBuffer, unsigned long *rxLength)
{  
  unsigned long i;

  if ( 0xFF < txLength )
    {
      std::cout << "max TLP packet length 255 exceeded -- TLPCLIB_SCardTransmit failed." << endl;
      return -1 ;
    }

  printf("[%03d]-> ", txLength);
  for (i=0; i < txLength; i++)
  {
	  printf("%02X ", txBuffer[i]);	
  }   printf("\n");

  cAPDU . set_APDU ( txBuffer , static_cast < Memory::size_t > ( txLength ) ) ;

#ifdef TLPCLIB_DEBUG
  std::cout << "calling the_tlp_apdu_server.exchange_APDUs ( cAPDU, rAPDU )\n";
#endif

  if ( 0 != the_tlp_apdu_server.exchange_APDUs ( cAPDU, rAPDU ) )
    {
      std::cout << "exchange_APDUs failed." << endl;
      return -1;
    };

#ifdef TLPCLIB_DEBUG
  std::cout << "returned from calling the_tlp_apdu_server.exchange_APDUs ( cAPDU, rAPDU )\n";
#endif

	*rxLength = static_cast < unsigned long > ( rAPDU . get_APDU_length ( ) ) ;

	for (i=0; i < *rxLength; i++)
	{
		rxBuffer[i] = rAPDU . get_APDU_octet( static_cast < Memory::size_t > ( i ) ) ;
	}

	printf("[%03d]<- ", *rxLength);
	for (i=0; i < *rxLength; i++)
	{
		printf("%02X ", rxBuffer[i]);
	}   printf("\n");


return 0;
}

int TLPCLIB_SCardReset()
{
#ifdef TLPCLIB_DEBUG
	std::cout << "resetting device" << endl;
#endif

return 0;
}
#if defined(_WIN32) && defined(__cplusplus)
extern "C" {
#endif  /* defined(_WIN32) && defined(__cplusplus) */

#include <string.h>
#include <hexify.h>


static char hexdigits[] = "0123456789ABCDEF";


BOOL
ishex (char c)
{
  return '\0' != c  &&  NULL != strchr (hexdigits, toupper (c));
}

int
hexvalue ( char c )
{
  const char *pdigit = strchr (hexdigits, toupper (c));
  return static_cast < int > 
           ( ( (char *) NULL == pdigit) ? -1 : pdigit - hexdigits ) ;
}

char
hexchar ( unsigned int i )
{
  return i < 16 ? hexdigits[i] : '?';
}

BOOL
isallhex (const char *p)
{
  for (; *p;)
    {
      if (!ishex (*p++))
        return false;
    }
  return true;
}


int
ahntol (const char *buf, const int nchars, unsigned long * pvalue)
{
  unsigned long val = 0;
  for (int n = 0; n < nchars; n++)
    {
      int h = hexvalue (buf[n]);
      if ( -1 == h )
        return n + 1;
      val = (val << 4) | h ;
    }
  * pvalue = val;
  return 0;
}


int
ahtol (const char *buf, unsigned long * pvalue)
{
  unsigned long val = 0;
  for (const char *p = buf; *p; p++)
    {
      int h = hexvalue (*p);
      if ( -1 == h )
        return static_cast < int > ( buf - p + 1 ) ;
      val = (val << 4) | h ;
    }
  * pvalue = val;
  return 0;
}


char *
ltoh ( unsigned long value , char * buf )
{
  char * p = buf + sizeof ( value ) ;
  for (int i = 0 ; i < sizeof ( value ) ; i ++ , value >>= 4 )
    * -- p  = hexdigits [ value & 0xF ] ;
  return buf ;
}


char *
lntoh ( unsigned long value , int n , char * buf )
{
  char * p = buf + n ;
  for (int i = 0 ; i < n ; i ++ , value >>= 4 )
    * -- p = hexdigits [ value & 0xF ] ;
  return buf ;
}


char *
bntoh ( unsigned char * b , int n , char * buf )
{
  char * p = buf ;
  for (int i = 0 ; i < n ; i ++ , b ++  )
    {
      * p ++ = hexdigits [ ( * b ) >> 4  ] ;
      * p ++ = hexdigits [ ( * b ) & 0xF ] ;
    }
  return buf ;
}


#if defined(_WIN32) && defined(__cplusplus)
} /* extern "C" */
#endif  /* defined(_WIN32) && defined(__cplusplus) */
