// Utility Progamming Macros
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__MACROS_H_INCLUDED__)
#define __MACROS_H_INCLUDED__

/*
** Alignment Macros
*/
#undef MAKEWORD
#undef MAKELONG
#undef MAKEDWORD
#undef HIDWORD
#undef LODWORD
#undef HIWORD
#undef LOWORD
#undef LOBYTE
#undef HIBYTE

#define MAKEWORD(hibyte, lobyte)      ((unsigned short int)(((unsigned char)(lobyte)) | ((unsigned short int)((unsigned char)(hibyte))) << 8))
#define MAKELONG(hiword, loword)      ((long int)(((unsigned short int)(loword)) | ((unsigned long int)((unsigned short int)(hiword))) << 16))
#define MAKEDWORD(hiword, loword)     ((unsigned long int)(((unsigned short int)(loword)) | ((unsigned long int)((unsigned short int)(hiword))) << 16))
// TODO: fix HIDWORD
#define HIDWORD(l)          ((unsigned long int)(0))
#define LODWORD(l)          ((unsigned long int)(l))
#define HIWORD(l)           ((unsigned short int)((unsigned long int)(l) >> 16))
#define LOWORD(l)           ((unsigned short int)(l))
#define LOBYTE(w)           ((unsigned char)((unsigned long int)(w) & 0xff))
#define HIBYTE(w)           ((unsigned char)((unsigned long int)(w) >> 8))

#define LITTLE_ENDIAN_WORD_BYTES(w) LOBYTE(w),HIBYTE(w)
#define LITTLE_ENDIAN_LOAD_WORD(p) MAKEWORD(1[p],0[p])
#define LITTLE_ENDIAN_STORE_WORD(p,w) (0[p]=LOBYTE(w),1[p]=HIBYTE(w),(w))
#define LITTLE_ENDIAN_STORE_WORD_AND_MOVE_POINTER(p,w) (LITTLE_ENDIAN_STORE_WORD((p),(w)),(p)+=2,(w))
#define LITTLE_ENDIAN_DWORD_BYTES(d) LITTLE_ENDIAN_WORD_BYTES(LOWORD(d)),LITTLE_ENDIAN_WORD_BYTES(HIWORD(d))
#define LITTLE_ENDIAN_LOAD_DWORD(p) MAKEDWORD(LITTLE_ENDIAN_LOAD_WORD((p)+2),LITTLE_ENDIAN_LOAD_WORD(p))
#define LITTLE_ENDIAN_STORE_DWORD(p,d) (LITTLE_ENDIAN_STORE_WORD((p),LOWORD(d)),LITTLE_ENDIAN_STORE_WORD(((p)+2),HIWORD(d)),(d))
#define LITTLE_ENDIAN_STORE_DWORD_AND_MOVE_POINTER(p,d) (LITTLE_ENDIAN_STORE_DWORD((p),(d)),(p)+=4,(d))
#define LITTLE_ENDIAN_TWORD_BYTES(t) LITTLE_ENDIAN_DWORD_BYTES(LODWORD(t)),LITTLE_ENDIAN_WORD_BYTES(LOWORD(HIDWORD(t)))
#define LITTLE_ENDIAN_LOAD_LODWORD_OF_QWORD(p) LITTLE_ENDIAN_LOAD_DWORD((p)+4)
#define LITTLE_ENDIAN_LOAD_HIDWORD_OF_QWORD(p) LITTLE_ENDIAN_LOAD_DWORD((p)+4)

#define BIG_ENDIAN_WORD_BYTES(w) HIBYTE(w),LOBYTE(w)
#define BIG_ENDIAN_LOAD_WORD(p) MAKEWORD(0[p],1[p])
#define BIG_ENDIAN_STORE_WORD(p,w) (0[p]=HIBYTE(w),1[p]=LOBYTE(w),(w))
#define BIG_ENDIAN_STORE_WORD_AND_MOVE_POINTER(p,w) (BIG_ENDIAN_STORE_WORD((p),(w)),(p)+=2,(w))
#define BIG_ENDIAN_DWORD_BYTES(d) BIG_ENDIAN_WORD_BYTES(HIWORD(d)),BIG_ENDIAN_WORD_BYTES(LOWORD(d))
#define BIG_ENDIAN_LOAD_DWORD(p) MAKEDWORD(BIG_ENDIAN_LOAD_WORD(p),BIG_ENDIAN_LOAD_WORD((p)+2))
#define BIG_ENDIAN_STORE_DWORD(p,d) (BIG_ENDIAN_STORE_WORD((p),HIWORD(d)),BIG_ENDIAN_STORE_WORD(((p)+2),LOWORD(d)),(d))
#define BIG_ENDIAN_STORE_DWORD_AND_MOVE_POINTER(p,d) (BIG_ENDIAN_STORE_DWORD((p),(d)),(p)+=4,(d))
#define BIG_ENDIAN_TWORD_BYTES(t) BIG_ENDIAN_WORD_BYTES(LOWORD(HIDWORD(t))),BIG_ENDIAN_DWORD_BYTES(LODWORD(t))
#define BIG_ENDIAN_LOAD_LODWORD_OF_QWORD(p) BIG_ENDIAN_LOAD_DWORD((p)+4)
#define BIG_ENDIAN_LOAD_HIDWORD_OF_QWORD(p) BIG_ENDIAN_LOAD_DWORD((p))

#if !defined(UNREFERENCED_PARAMETER)
  #define UNREFERENCED_PARAMETER(P) (P)
#endif


#if !defined(LITTLE_ENDIAN)
  #define LITTLE_ENDIAN 1234
#endif

#if !defined(BIG_ENDIAN)
  #define BIG_ENDIAN 4321
#endif

#if !defined(BYTE_ORDER)
   #if defined(_WIN32) || defined(_CYGWIN)              /* or other little-endians */
      #define BYTE_ORDER LITTLE_ENDIAN
   #elif 0                         /* or other big-endians */
      #define BYTE_ORDER BIG_ENDIAN
   #else
      #error Must define endianness for this build
   #endif
#endif   // !defined(BYTE_ORDER)

#define NETWORK_BYTE_ORDER BIG_ENDIAN

#if BYTE_ORDER==LITTLE_ENDIAN

#define HOST_WORD_BYTES                       LITTLE_ENDIAN_WORD_BYTES
#define HOST_LOAD_WORD                        LITTLE_ENDIAN_LOAD_WORD
#define HOST_STORE_WORD                       LITTLE_ENDIAN_STORE_WORD
#define HOST_STORE_WORD_AND_MOVE_POINTER      LITTLE_ENDIAN_STORE_WORD_AND_MOVE_POINTER
#define HOST_DWORD_BYTES                      LITTLE_ENDIAN_DWORD_BYTES
#define HOST_LOAD_DWORD                       LITTLE_ENDIAN_LOAD_DWORD
#define HOST_STORE_DWORD                      LITTLE_ENDIAN_STORE_DWORD
#define HOST_STORE_DWORD_AND_MOVE_POINTER     LITTLE_ENDIAN_STORE_DWORD_AND_MOVE_POINTER
#define HOST_TWORD_BYTES                      LITTLE_ENDIAN_TWORD_BYTES
#define HOST_LOAD_LODWORD_OF_QWORD            LITTLE_ENDIAN_LOAD_LODWORD_OF_QWORD
#define HOST_LOAD_HIDWORD_OF_QWORD            LITTLE_ENDIAN_LOAD_HIDWORD_OF_QWORD
                                                
#else // !BYTE_ORDER==LITTLE_ENDIAN
                                                
#define HOST_WORD_BYTES                       BIG_ENDIAN_WORD_BYTES
#define HOST_LOAD_WORD                        BIG_ENDIAN_LOAD_WORD
#define HOST_STORE_WORD                       BIG_ENDIAN_STORE_WORD
#define HOST_STORE_WORD_AND_MOVE_POINTER      BIG_ENDIAN_STORE_WORD_AND_MOVE_POINTER
#define HOST_DWORD_BYTES                      BIG_ENDIAN_DWORD_BYTES
#define HOST_LOAD_DWORD                       BIG_ENDIAN_LOAD_DWORD
#define HOST_STORE_DWORD                      BIG_ENDIAN_STORE_DWORD
#define HOST_STORE_DWORD_AND_MOVE_POINTER     BIG_ENDIAN_STORE_DWORD_AND_MOVE_POINTER
#define HOST_TWORD_BYTES                      BIG_ENDIAN_TWORD_BYTES
#define HOST_LOAD_LODWORD_OF_QWORD            BIG_ENDIAN_LOAD_LODWORD_OF_QWORD
#define HOST_LOAD_HIDWORD_OF_QWORD            BIG_ENDIAN_LOAD_HIDWORD_OF_QWORD

#endif /* LITTLE_ENDIAN */


#define NETWORK_WORD_BYTES                    BIG_ENDIAN_WORD_BYTES
#define NETWORK_LOAD_WORD                     BIG_ENDIAN_LOAD_WORD
#define NETWORK_STORE_WORD                    BIG_ENDIAN_STORE_WORD
#define NETWORK_STORE_WORD_AND_MOVE_POINTER   BIG_ENDIAN_STORE_WORD_AND_MOVE_POINTER
#define NETWORK_DWORD_BYTES                   BIG_ENDIAN_DWORD_BYTES
#define NETWORK_LOAD_DWORD                    BIG_ENDIAN_LOAD_DWORD
#define NETWORK_STORE_DWORD_AND_MOVE_POINTER  BIG_ENDIAN_STORE_DWORD_AND_MOVE_POINTER
#define NETWORK_TWORD_BYTES                   BIG_ENDIAN_TWORD_BYTES
#define NETWORK_LOAD_LODWORD_OF_QWORD         BIG_ENDIAN_LOAD_LODWORD_OF_QWORD
#define NETWORK_LOAD_HIDWORD_OF_QWORD         BIG_ENDIAN_LOAD_HIDWORD_OF_QWORD

#endif // !defined(__MACROS_H_INCLUDED__)
