// Status Words
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__SW_H_INCLUDED__)
#define __SW_H_INCLUDED__



/* Error checking and filtering macros */

#if defined(DEBUG_BUILD)
#define OK_or_die(scexp,sw) \
  do \
    { \
      register const SCODE __sc = (scexp); \
      if ( FAILED(__sc) ) \
        { \
          PRINTF(("%s(%d): failed with SCODE %08X, returning error SW %04X\n", \
                 __FILE__,__LINE__,__sc,__sw); \
          return ((unsigned short int)(sw)) ; \
        } \
    } \
  while ( 0 )

#define SW_OK_or_die(swexp) \
  do \
    { \
      register const unsigned short int __sw = (swexp); \
      if ( SW_OK != __sw ) \
        { \
          PRINTF(("%s(%d): failed, returning error SW %04X\n", __FILE__,__LINE__,__sw); \
          return __sw ; \
        } \
    } \
  while ( 0 )
#else  /* defined(DEBUG_BUILD) */
#define OK_or_die(sc,sw) do if ( FAILED(sc) ) return ((unsigned short int)sw) ; while ( 0 )
#define SW_OK_or_die(swexp) do { register const unsigned short int __sw = (swexp); if ( SW_OK != __sw ) return __sw ; } while ( 0 )
#endif /* defined(DEBUG_BUILD) */

#define OK_or_die_MP(sc)    OK_or_die( (sc) , (unsigned short int)SW_MEMORY_PROBLEM )
#define OK_or_die_FNF(sc)   OK_or_die( (sc) , (unsigned short int)SW_FILE_NOT_FOUND )
#define OK_or_die_SSNS(sc)  OK_or_die( (sc) , (unsigned short int)SW_SECURITY_STATUS_NOT_SATISFIED )

#define OK_or_sw(sc,sw)        ( FAILED(sc) ? ((unsigned short int)sw) : (unsigned short int)SW_OK )
#define OK_or_MP(sc)           OK_or_sw(sc, (unsigned short int)SW_MEMORY_PROBLEM)


#endif /* !defined(__SW_H_INCLUDED__) */
