#include "stdafx.h"
#include "EventLog.h"

BOOL AddEventSource(
   LPCTSTR pszLogName, // Application log or a custom log
   LPCTSTR pszSrcName, // event source name
   LPCTSTR pszMsgDLL,  // path for message DLL
   DWORD  dwNum)      // number of categories
{
   HKEY hk;
   DWORD dwData, dwDisp;
   TCHAR szBuf[MAX_PATH];

   // Create the event source as a subkey of the log. 
 
   wsprintf(szBuf, 
      "SYSTEM\\CurrentControlSet\\Services\\EventLog\\%s\\%s",
      pszLogName, pszSrcName); 
 
   if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, szBuf, 
          0, NULL, REG_OPTION_NON_VOLATILE,
          KEY_WRITE, NULL, &hk, &dwDisp)) 
   {
      printf("Could not create the registry key."); 
      return FALSE;
   }
 
   // Set the name of the message file. 
 
   if (RegSetValueEx(hk,             // subkey handle 
          "EventMessageFile",        // value name 
          0,                         // must be zero 
          REG_EXPAND_SZ,             // value type 
          (LPBYTE) pszMsgDLL,        // pointer to value data 
          (DWORD) (lstrlen(pszMsgDLL)+1)*sizeof(TCHAR))) // data size
   {
      printf("Could not set the event message file."); 
      RegCloseKey(hk); 
      return FALSE;
   }
 
   // Set the supported event types. 
 
   dwData = EVENTLOG_SUCCESS | EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | 
        EVENTLOG_INFORMATION_TYPE; 
 
   if (RegSetValueEx(hk,      // subkey handle 
           "TypesSupported",  // value name 
           0,                 // must be zero 
           REG_DWORD,         // value type 
           (LPBYTE) &dwData,  // pointer to value data 
           sizeof(DWORD)))    // length of value data 
   {
      printf("Could not set the supported types."); 
      RegCloseKey(hk); 
      return FALSE;
   }
 
   // Set the category message file and number of categories.

   if (RegSetValueEx(hk,              // subkey handle 
           "CategoryMessageFile",     // value name 
           0,                         // must be zero 
           REG_EXPAND_SZ,             // value type 
           (LPBYTE) pszMsgDLL,        // pointer to value data 
           (DWORD) (lstrlen(pszMsgDLL)+1)*sizeof(TCHAR))) // data size
   {
      printf("Could not set the category message file."); 
      RegCloseKey(hk); 
      return FALSE;
   }
 
   if (RegSetValueEx(hk,      // subkey handle 
           "CategoryCount",   // value name 
           0,                 // must be zero 
           REG_DWORD,         // value type 
           (LPBYTE) &dwNum,   // pointer to value data 
           sizeof(DWORD)))    // length of value data 
   {
      printf("Could not set the category count."); 
      RegCloseKey(hk); 
      return FALSE;
   }

   RegCloseKey(hk); 
   return TRUE;
}

void MyReportEvent(
    LPCTSTR pszSrcName, // event source name
	WORD wType,        // event type
    DWORD dwEventID,   // event identifier
    WORD wCategory,    // event category
    WORD cInserts,     // count of insert strings
    LPCTSTR *szMsg)    // insert strings
{
    HANDLE h; 

    // Get a handle to the event log.
 
    h = RegisterEventSource(NULL,  // use local computer 
            pszSrcName);           // event source name 
    if (h == NULL) 
    {
        printf("Could not register the event source."); 
        return;
    }

    // Report the event.
 
    if (!ReportEvent(h,           // event log handle 
            wType,                // event type 
            wCategory,            // event category  
            dwEventID,            // event identifier 
            NULL,                 // no user security identifier 
            cInserts,             // number of substitution strings 
            0,                    // no data 
            szMsg,                // pointer to strings 
            NULL))                // no data 
    {
        printf("Could not report the event."); 
    }
 
    DeregisterEventSource(h); 
    return;
}

BOOL AddDefaultEventSource(LPCTSTR pszSourceName)
{
	// get the path to the application
	CString cStrApplicationPath = ::GetCommandLine();
	int iPathLength = cStrApplicationPath.ReverseFind('\\');
	TCHAR szBuf[MAX_PATH];
	wsprintf(szBuf, "%s\\%s", cStrApplicationPath.Mid(0, iPathLength).TrimLeft('"'), pcszMsgDLLName); 

	return AddEventSource(pcszLogName, pszSourceName, szBuf, dwCategoryCount);
}

void ReportDefaultSuccessEvent(LPCTSTR pszSourceName, LPCTSTR szMsg)
{
	LPCTSTR * pszMsg = &szMsg;
	return MyReportEvent(pszSourceName, EVENTLOG_SUCCESS, MSG_SUCCESS, CATEGORY1, 1, pszMsg);
}

void ReportDefaultInformationEvent(LPCTSTR pszSourceName, LPCTSTR szMsg)
{
	LPCTSTR * pszMsg = &szMsg;
	return MyReportEvent(pszSourceName, EVENTLOG_INFORMATION_TYPE, MSG_INFO, CATEGORY1, 1, pszMsg);
}

void ReportDefaultWarningEvent(LPCTSTR pszSourceName, LPCTSTR szMsg)
{
	LPCTSTR * pszMsg = &szMsg;
	return MyReportEvent(pszSourceName, EVENTLOG_WARNING_TYPE, MSG_WARN, CATEGORY1, 1, pszMsg);
}

void ReportDefaultErrorEvent(LPCTSTR pszSourceName, LPCTSTR szMsg)
{
	LPCTSTR * pszMsg = &szMsg;
	return MyReportEvent(pszSourceName, EVENTLOG_ERROR_TYPE, MSG_ERROR, CATEGORY1, 1, pszMsg);
}