#include "stdafx.h"
#include "Winioctl.h"
#include "PIV VSCR Service.h"
#include "WDMIoctl.h"
#include "EventThread.h"

// name of VSCR card events section
#define VSCR_EVENTS_SECTION				"VSCR Card Events"
#define VSCR_EVENT_COUNT_KEYNAME		"EventCount"


//this function will be called when the thread startup
DWORD WINAPI EventThread(LPVOID lpParameter)
{
	CEventThread* pEventThread = (CEventThread*) lpParameter;
	EVENT_INFO Event;
	HANDLE hVscrDriver;
	CTime ctCurrTime;
	POSITION Pos, CurrPos;
	DWORD dwIoctl, dwBytesReturned;

	pEventThread->m_bStopping = FALSE;

	while (!pEventThread->m_bStopping) {

		// get current time
		ctCurrTime = CTime::GetCurrentTime();

		Pos = pEventThread->m_clEvents.GetHeadPosition();

		// find event that matches the current time
		while (Pos != NULL) {

			CurrPos = Pos;
			Event = pEventThread->m_clEvents.GetNext(Pos);

			if ((Event.dHour == ctCurrTime.GetHour()) &&
				(Event.dMinute == ctCurrTime.GetMinute()) &&
				(Event.dSecond == ctCurrTime.GetSecond())) {

				// get handle to VSCR driver
				hVscrDriver = CreateFile (
									PIV_VSCRDRV_DEVICE_NAME,
									GENERIC_ALL,
									FILE_SHARE_READ,
									NULL,
									OPEN_EXISTING,
									0,
									NULL);

				if (hVscrDriver != INVALID_HANDLE_VALUE) {

					if (Event.eType == CARD_INSERTED)
						dwIoctl = IOCTL_SMARTCARD_INSERT;
					else
						dwIoctl = IOCTL_SMARTCARD_REMOVE;

					// send IOCTL to VSCR driver
					DeviceIoControl(
							hVscrDriver,
							dwIoctl,
							NULL,
							0,
							NULL,
							0,
							&dwBytesReturned,
							NULL);

					CloseHandle(hVscrDriver);
				}

				// remove event from list so we don't send the IOCTL again
				pEventThread->m_clEvents.RemoveAt(CurrPos);

				break;
			}
		}

		// wait 250 ms before next execution of thread
		Sleep(250);
	}

	return 0;
}

CEventThread::CEventThread()
{
	m_hThread = NULL;
}

CEventThread::~CEventThread()
{
	// if the thread is running then stop it
	Stop();
}

DWORD CEventThread::Start()
{
	DWORD dwStatus = ERROR_FUNCTION_FAILED;

	// read events
	ReadEvents();

	// create thread to process messages
	m_hThread = CreateThread(
						NULL,
						0, 
						(LPTHREAD_START_ROUTINE)EventThread, 
						(LPVOID)this, 
						0, 
						NULL);

	if (m_hThread)
		dwStatus = ERROR_SUCCESS;
	else
		m_clEvents.RemoveAll();

	return dwStatus;
}

void CEventThread::Stop(void)
{
	// stop our thread
	if (m_hThread) {

		// set flag indicating the thread should be stopped
		m_bStopping = TRUE;

		// Clearing m_bStopping will cause the thread to stop.  If it hasn't
		// stopped then give it a second before we terminate the thread.
		if (WaitForSingleObject(m_hThread, 1000) == WAIT_TIMEOUT)
			TerminateThread(m_hThread, -1);

		CloseHandle(m_hThread);
		m_hThread = NULL;
	}

	// cleanup event list
	m_clEvents.RemoveAll();
}

void CEventThread::ReadEvents()
{
	EVENT_INFO Event;
	char szSysDir[MAX_PATH+1], szEventIniFile[MAX_PATH];
	char szKeyName[32], szValue[32];
	int dEventCount;

	m_clEvents.RemoveAll();

	// get name of Windows system directory
	GetSystemDirectory(szSysDir, sizeof(szSysDir));

	if (szSysDir[strlen(szSysDir)-1] != '\\')
		strcat(szSysDir, "\\");

	sprintf(szEventIniFile, "%s%s", szSysDir, "VSCR Event Script.ini");

	// get count of events
	dEventCount = GetPrivateProfileInt(
						VSCR_EVENTS_SECTION,
						VSCR_EVENT_COUNT_KEYNAME,
						0,
						szEventIniFile);


	// read events
	for (int x=0; x < dEventCount; x++) {

		sprintf(szKeyName, "Event%dType", x+1);

		// get event type
		GetPrivateProfileString(
				VSCR_EVENTS_SECTION,
				szKeyName,
				"NULL",
				szValue,
				sizeof(szValue),
				szEventIniFile);

		if (stricmp(szValue, "Card Insertion") == 0)
			Event.eType = CARD_INSERTED;
		else if (stricmp(szValue, "Card Removal") == 0)
			Event.eType = CARD_REMOVED;
		else
			continue;

		sprintf(szKeyName, "Event%dTime", x+1);

		// get event type
		GetPrivateProfileString(
				VSCR_EVENTS_SECTION,
				szKeyName,
				"NULL",
				szValue,
				sizeof(szValue),
				szEventIniFile);

		if (strcmp(szValue, "NULL") != 0) {

			sscanf(szValue, "%02d:%02d:%02d", &Event.dHour, &Event.dMinute, 
				&Event.dSecond);

			// add event to list
			m_clEvents.AddTail(Event);
		}
	}
}